from __future__ import division, unicode_literals, print_function
import base64
import copy
import gzip
import math
import os
import operator
import traceback
import random
import signal
import sys
from collections import defaultdict, OrderedDict
from functools import partial
from multiprocessing import Queue, Manager
from queue import Empty
from string import Template

import pandas as pd
from pythomics.proteomics.config import CARBON_NEUTRON, PROTON
from pythomics.proteomics.parsers import GuessIterator
from pythomics.proteomics import config
from scipy.interpolate import UnivariateSpline

from .reader import Reader
from .worker import Worker
from .utils import (
    find_prior_scan,
    get_scans_under_peaks,
    naninfmean,
    naninfsum,
    perform_ml,
    get_formatted_mass,
)
from . import peaks
from pyquant.cpeaks import find_nearest_indices


description = """
PyQuant is a quantification program for mass spectrometry data. It attempts to be a general implementation to quantify
an assortment of datatypes and allows a high degree of customization for how data is to be quantified.
"""

ION_CUTOFF = 2

CRASH_SIGNALS = {
    signal.SIGSEGV,
}


def run_pyquant():
    from . import pyquant_parser

    if len(sys.argv) == 1:
        pyquant_parser.print_help()
        sys.exit()

    # for some reason this is being undefined in windows
    import numpy as np

    args = pyquant_parser.parse_args()
    isotopologue_limit = args.isotopologue_limit
    isotopologue_limit = isotopologue_limit if isotopologue_limit else None
    labels_needed = args.labels_needed
    overlapping_mz = args.overlapping_labels
    threads = args.p
    skip = args.skip
    out = args.out
    html = args.html
    resume = args.resume
    calc_stats = not args.disable_stats
    msn_for_id = args.msn_id
    mass_accuracy_correction = args.no_mass_accuracy_correction
    raw_data_only = not (args.search_file or args.tsv)
    scans_to_select = set(args.scan if args.scan else [])
    msn_for_quant = args.msn_quant_from if args.msn_quant_from else msn_for_id - 1
    if msn_for_quant == 0:
        msn_for_quant = 1
    reporter_mode = args.reporter_ion
    msn_ppm = args.msn_ppm
    if args.msn_rt_window:
        msn_rt_window = [tuple(map(float, i.split("-"))) for i in args.msn_rt_window]
    else:
        msn_rt_window = None
    ref_label = str(args.reference_label) if args.reference_label else None
    quant_method = args.quant_method
    if msn_for_quant == 1 and quant_method is None:
        quant_method = "integrate"
    elif msn_for_quant > 1 and quant_method is None:
        quant_method = "sum"

    if args.isobaric_tags:
        reporter_mode = True
        mass_accuracy_correction = False
        calc_stats = False
        msn_for_quant = 2
    if args.ms3:
        msn_for_quant = 3
        mass_accuracy_correction = False
    if args.gcms:
        msn_for_quant = 1
        msn_for_id = 1
        args.precursor_ppm = args.msn_ppm
        isotopologue_limit = 1
        args.msn_all_scans = True
        args.no_mass_accuracy_correction = True
        args.no_contaminant_detection = True
        args.no_rt_guide = True
        args.disable_stats = True

    mrm_pair_info = (
        pd.read_csv(args.mrm_map, sep="\t") if args.mrm and args.mrm_map else None
    )

    scan_filemap = {}
    found_scans = {}
    raw_files = {}
    mass_labels = {}  # {'Light': {0: set([])}} if not reporter_mode else {}

    name_mapping = {}

    if args.label_scheme:
        label_info = pd.read_csv(
            args.label_scheme.name, sep="\t", header=None, dtype="str"
        )
        try:
            label_info.columns = ["Label", "AA", "Mass", "UserName"]
            name_mapping = dict(
                [(v["Label"], v["UserName"]) for i, v in label_info.iterrows()]
            )
        except ValueError:
            label_info.columns = ["Label", "AA", "Mass"]
            name_mapping = dict(
                [(v["Label"], v["Label"]) for i, v in label_info.iterrows()]
            )
        for group_name, group_info in label_info.groupby("Label"):
            masses = {}
            label_name = name_mapping.get(group_name, group_name)
            for mass, mass_aas in group_info.groupby("Mass"):
                mass_val = float(mass)
                mass_list = mass_aas["AA"].values.tolist()
                try:
                    masses[mass_val].add(mass_list)
                except KeyError:
                    masses[mass_val] = set(mass_list)
            mass_labels.update({label_name: masses})
    if args.label_method:
        mass_labels = config.LABEL_SCHEMES[args.label_method]

    sample = args.sample
    sys.stderr.write("Loading Scans:\n")

    # options determining modes to quantify
    all_msn = False  # we just have a raw file
    ion_search = False  # we have an ion we want to find

    if args.peptide_file:
        peptides = set((i.strip().upper() for i in args.peptide_file))
    elif args.peptide:
        peptides = (
            set(args.peptide) if isinstance(args.peptide, list) else set([args.peptide])
        )
    else:
        peptides = None

    if peptides:
        peptides = set(map(lambda x: x.upper(), peptides))

    input_found = None
    if args.search_file:
        results = GuessIterator(
            args.search_file.name, full=True, store=False, peptide=peptides
        )
        input_found = "ms"
    elif args.tsv:
        results = pd.read_csv(args.tsv, sep="\t")
        input_found = "tsv"

    if args.search_file:
        source_file = args.search_file.name
    elif args.tsv:
        source_file = args.tsv.name
    elif args.scan_file:
        source_file = args.scan_file[0].name
    elif args.scan_file_dir:
        source_file = os.path.split(args.scan_file_dir)[1]
    else:
        pyquant_parser.error(
            "No input files provided. Please provide one via --search-file, --tsv, --scan-file, or --scan-file-dir."
        )

    if args.scan_file:
        nfunc = (
            lambda i: (
                os.path.splitext(os.path.split(i.name)[1])[0],
                os.path.abspath(i.name),
            )
            if hasattr(i, "name")
            else (os.path.splitext(os.path.split(i)[1])[0], os.path.abspath(i))
        )
        scan_filemap = dict([nfunc(i) for i in args.scan_file])
    else:
        if args.scan_file_dir:
            raw_file = args.scan_file_dir
        else:
            raw_file = os.path.abspath(os.path.split(source_file)[0])
        if os.path.isdir(raw_file):
            scan_filemap = dict(
                [
                    (os.path.splitext(i)[0], os.path.abspath(os.path.join(raw_file, i)))
                    for i in os.listdir(raw_file)
                    if i.lower().endswith("mzml")
                ]
            )
        else:
            scan_filemap[
                os.path.splitext(os.path.split(raw_file)[1])[0]
            ] = os.path.abspath(raw_file)

    result_iterator = []
    replicate_file_mapper = {}

    if input_found == "tsv":
        if args.maxquant:
            peptide_col = "Sequence"
            precursor_col = "Mass"
            rt_col = "Retention time"
            charge_col = "Charge"
            file_col = "Raw file"
            scan_col = "MS/MS Scan Number"
            label_col = "Labeling State"
            if "ms2" in source_file:
                scan_col = "Scan number"
                label_col = None
        else:
            peptide_col = args.peptide_col
            precursor_col = args.mz
            rt_col = args.rt
            charge_col = args.charge
            file_col = args.source
            label_col = args.label
            scan_col = args.scan_col

        result_iterator = results.iterrows()

        def tsv_formatter(row):
            row_index, row_info = row
            peptide = row_info[peptide_col].strip() if peptide_col in row_info else ""
            specId = str(row_info[scan_col])
            fname = row_info[file_col] if file_col in row_info else raw_file
            charge = float(row_info[charge_col]) if charge_col in row_info else 1
            precursor_mass = (
                row_info[precursor_col] if precursor_col in row_info else None
            )
            rt_value = row_info[rt_col] if rt_col in row_info else None
            label = (
                name_mapping.get(row_info[label_col]) if label_col in row_info else None
            )

            if math.isnan(precursor_mass) or not charge:
                return None

            if args.maxquant:
                precursor_mass = precursor_mass / charge + PROTON

            return {
                "file": fname,
                "id": specId,
                "mass": float(precursor_mass),
                "theor_mass": float(precursor_mass),
                "rt": rt_value,
                "charge": charge,
                "peptide": peptide,
                "modifications": None,
                "label": label,
                "key": (fname, specId, peptide, precursor_mass),
            }

        result_formatter = tsv_formatter

    elif input_found == "ms":
        if not (args.label_scheme or args.label_method):
            mass_labels.update(results.getSILACLabels())
        result_iterator = results.getScans(modifications=False, fdr=True)

        def ms_formatter(scan):
            peptide = scan.peptide
            specId = scan.id
            fname = scan.file

            return {
                "file": fname,
                "id": specId,
                "mass": float(scan.mass),
                "rt": scan.rt,
                "charge": scan.charge,
                "peptide": peptide,
                "mod_peptide": scan.modifiedPeptide,
                "theor_mass": scan.getTheorMass(),
                "modifications": scan.getModifications(),
                "accession": getattr(scan, "acc", None),
                "key": (fname, specId, peptide, scan.mass),
            }

        result_formatter = ms_formatter

    for index, scan in enumerate(result_iterator):
        if index % 1000 == 0:
            sys.stderr.write(".")
        if scan is None:
            continue

        result_info = result_formatter(scan)
        if result_info is None:
            continue

        peptide = result_info["peptide"]

        if peptides and peptide.upper() not in peptides:
            continue
        if not peptides and (sample != 1.0 and random.random() > sample):
            continue

        specId = result_info["id"]
        if scans_to_select and str(specId) not in scans_to_select:
            continue

        fname = result_info["file"]
        mass_key = result_info["key"]
        if mass_key in found_scans:
            continue

        d = {
            "file": fname,
            "quant_scan": {},
            "id_scan": {
                "id": specId,
                "theor_mass": result_info.get("theor_mass"),
                "peptide": peptide,
                "mod_peptide": result_info.get("mod_peptide"),
                "rt": result_info["rt"],
                "charge": result_info["charge"],
                "modifications": result_info["modifications"],
                "mass": result_info["mass"],
                "accession": result_info.get("accession"),
            },
        }
        found_scans[mass_key] = d
        fname = os.path.splitext(fname)[0]
        if args.mva:
            # find the most similar name, add in a setting for this
            import difflib

            if fname in replicate_file_mapper:
                fname = replicate_file_mapper[fname]
                if fname is None:
                    continue
            else:
                s = difflib.SequenceMatcher(None, fname)
                seq_matches = []
                for i in scan_filemap:
                    s.set_seq2(i)
                    seq_matches.append((s.ratio(), i))
                seq_matches = [
                    i
                    for i in sorted(
                        seq_matches, key=operator.itemgetter(0), reverse=True
                    )
                    if i[0] != 1
                ]
                replicate_file_mapper[fname] = seq_matches[0][1]
                fname = seq_matches[0][1]
        if fname not in scan_filemap:
            fname = os.path.split(fname)[1]
            if fname not in scan_filemap:
                if skip:
                    continue
                sys.stderr.write(
                    "{0} not found in filemap. Filemap is {1}.".format(
                        fname, scan_filemap
                    )
                )
                return 1
        try:
            raw_files[fname].append(d)
        except KeyError:
            raw_files[fname] = [d]
        del scan

    if scan_filemap and raw_data_only:
        # determine if we want to do ms1 ion detection, ms2 ion detection, all ms2 of each file
        if args.msn_ion or args.msn_peaklist:
            ion_search = True
            ions_of_interest = []
            if args.msn_peaklist:
                for i in args.msn_peaklist:
                    if not i.strip():
                        continue
                    try:
                        ions_of_interest.append(i.strip())
                    except Exception as e:
                        traceback.print_exc()
            if args.msn_ion:
                ions_of_interest += args.msn_ion
            ions_selected = (
                [
                    sorted(list(set(map(float, ion.split(",")))))
                    for ion in ions_of_interest
                ]
                if ions_of_interest
                else []
            )
            d = {"ions": ions_selected, "rt_info": args.msn_ion_rt}
            for i in scan_filemap:
                raw_files[i] = d
        else:
            all_msn = True
            for i in scan_filemap:
                raw_files[i] = [1]
    if not scan_filemap and input_found is None:
        sys.stderr.write(
            "No valid input entered. PyQuant requires at least a raw file or a processed dataset."
        )
        return 1
    sys.stderr.write("\nScans loaded.\n")

    pq_dir = os.path.split(__file__)[0]

    pyquant_html_file = os.path.join(pq_dir, "static", "pyquant_output.html")

    workers = []
    completed = 0
    sys.stderr.write("Beginning quantification.\n")
    scan_count = len(found_scans)

    # before we start, get the information we will be generating
    labels = list(mass_labels.keys())
    if not labels:
        if ion_search:
            if args.require_all_ions:
                labels = ["_".join(map(str, ion_set)) for ion_set in ions_selected]
            else:
                labels = list(
                    map(str, set([ion for ion_set in ions_selected for ion in ion_set]))
                )
        else:
            labels = [""]
    if not scan_filemap and raw_data_only:
        RESULT_ORDER = []
    else:
        RESULT_ORDER = [
            ("peptide", "Peptide"),
            ("modifications", "Modifications"),
            ("accession", "Accession"),
        ]
    RESULT_ORDER.extend(
        [("charge", "Charge"), ("ms1", "MS{} Spectrum ID".format(msn_for_quant)),]
    )
    if msn_for_quant != msn_for_id:
        RESULT_ORDER.extend(
            [("scan", "MS{} Spectrum ID".format(msn_for_id)),]
        )
    RESULT_ORDER.extend(
        [("rt", "Retention Time"),]
    )

    if ion_search and not args.msn_all_scans:
        RESULT_ORDER.extend([("ions_found", "Ions Found")])

    PEAK_REPORTING = []
    if not reporter_mode and args.peaks_n != 1:
        PEAK_REPORTING.extend(
            [
                ("label", "Peak Label"),
                ("auc", "Peak Area"),
                ("amp", "Peak Max"),
                ("peak_mean", "Peak Center"),
                ("peak_width", "RT Width"),
                ("mean_diff", "Mean Offset"),
                ("snr", "SNR"),
                ("sbr", "SBR"),
                ("sdr", "Density"),
                ("coef_det", "R^2"),
                ("residual", "Residual"),
            ]
        )
    iterator = ["_".join(map(str, sorted(labels)))] if args.merge_labels else labels
    for label in iterator:
        label_key = "{} ".format(label) if label != "" else label
        RESULT_ORDER.extend(
            [("{}_precursor".format(label), "{}Precursor".format(label_key)),]
        )
        if msn_for_quant == 1:
            if not args.no_mass_accuracy_correction:
                RESULT_ORDER.extend(
                    [
                        (
                            "{}_calibrated_precursor".format(label),
                            "{}Calibrated Precursor".format(label_key),
                        ),
                    ]
                )
            RESULT_ORDER.extend(
                [("{}_isotopes".format(label), "{}Isotopes Found".format(label_key)),]
            )
        RESULT_ORDER.extend(
            [
                (
                    "{}_intensity".format(label),
                    "{}Intensity".format(label_key),
                    partial(naninfsum, empty=0),
                ),
            ]
        )
        if args.peaks_n == 1:
            RESULT_ORDER.extend(
                [
                    (
                        "{}_peak_width".format(label),
                        "{}RT Width".format(label_key),
                        partial(naninfmean, empty=0),
                    ),
                    (
                        "{}_peak_mean".format(label),
                        "{}Peak Center".format(label_key),
                        partial(naninfmean, empty=np.NaN),
                    ),
                    (
                        "{}_mean_diff".format(label),
                        "{}Mean Offset".format(label_key),
                        partial(naninfmean, empty=np.NaN),
                    ),
                ]
            )

        if msn_for_quant == 1:
            RESULT_ORDER.extend(
                [
                    (
                        "{}_residual".format(label),
                        "{}Residual".format(label_key),
                        partial(naninfsum, empty=np.NaN),
                    ),
                    (
                        "{}_coef_det".format(label),
                        "{}R^2".format(label_key),
                        partial(naninfmean, empty=0),
                    ),
                    (
                        "{}_snr".format(label),
                        "{}SNR".format(label_key),
                        partial(naninfmean, empty=0),
                    ),
                ]
            )
        if not args.merge_labels and not args.no_ratios and label != "":
            for label2 in labels:
                label_key2 = "{}".format(label2) if label2 != "" else label2
                if label_key != label_key2 and (
                    ref_label is None or ref_label.lower() == label_key2.lower()
                ):
                    RESULT_ORDER.extend(
                        [
                            (
                                "{}_{}_ratio".format(label, label2),
                                "{}/{}".format(label_key.strip(), label_key2.strip()),
                            ),
                        ]
                    )
                    if calc_stats:
                        RESULT_ORDER.extend(
                            [
                                (
                                    "{}_{}_confidence".format(label, label2),
                                    "{}/{} Confidence".format(
                                        label_key.strip(), label_key2.strip()
                                    ),
                                ),
                            ]
                        )

    headers = ["Raw File"] + [i[1] for i in RESULT_ORDER]
    if resume and os.path.exists(out):
        if not out:
            sys.stderr.write("You may only resume runs with a file output.\n")
            return -1
        out = open(out, "a+")
        out_path = out.name
    else:
        if out:
            out = open(out, "w+")
            out_path = out.name
        else:
            out = sys.stdout
            out_path = source_file
        out.write("{0}\n".format("\t".join(headers)))

    if html:

        def table_rows(html_list, res=None):
            # each item is a string like a\tb\tc
            if html_list:
                d = html_list.pop(0)
            else:
                return res
            line = d["table"]
            html_extra = d.get("html", {})
            keys = d["keys"]
            if res is None:
                res = "<tr>"
            out = ["<td></td>"]  # for graph controls
            for col_index, (i, v) in enumerate(zip(line.split("\t"), keys)):
                out.append("<td>{0}</td>".format(i))
            res += "\n".join(out) + "</tr>"
            return table_rows(html_list, res=res), html_extra

        if resume:
            html_out = open("{0}.html".format(out_path), "a")
        else:
            html_out = open("{0}.html".format(out_path), "w")
            template = []
            for i in open(pyquant_html_file, "r"):
                if "HTML BREAK" in i:
                    break
                template.append(i)
            html_template = Template("".join(template))
            html_out.write(
                html_template.substitute(
                    {
                        "title": source_file,
                        "table_header": "\n".join(
                            [
                                "<th>{0}</th>".format(i)
                                for i in ["Controls", "Raw File"]
                                + [i[1] for i in RESULT_ORDER]
                            ]
                        ),
                    }
                )
            )

    skip_map = set([])
    if resume:
        resume_name = "{}.tmp".format(out.name)
        if os.path.exists(resume_name):
            with open(resume_name, "r") as temp_file:
                for index, entry in enumerate(temp_file):
                    info = pd.io.json.loads(entry)["res_list"]
                    # key is filename, peptide, charge, target scan id, modifications
                    key = tuple(map(str, (info[0], info[1], info[3], info[5], info[2])))
                    skip_map.add(key)
            temp_file = open(resume_name, "a")
        else:
            temp_file = open(resume_name, "w")
    else:
        temp_file = open("{}.tmp".format(out.name), "w")

    silac_shifts = {}
    for silac_label, silac_masses in mass_labels.items():
        for mass, aas in silac_masses.items():
            try:
                silac_shifts[get_formatted_mass(mass)] |= aas
            except Exception as e:
                silac_shifts[get_formatted_mass(mass)] = aas

    for filename in raw_files.keys():
        raw_scans = raw_files[filename]
        filepath = scan_filemap[filename]
        if not len(raw_scans):
            continue
        in_queue = Queue()
        result_queue = Queue()
        reader_in = Queue()
        reader_outs = {"main": Queue()}
        for i in range(threads):
            reader_outs[i] = Queue()

        msn_map = []
        scan_rt_map = {}
        msn_rt_map = {}
        scan_charge_map = {}

        raw = GuessIterator(filepath, full=False, store=False)
        sys.stderr.write("Processing {}.\n".format(filepath))

        # params in case we are doing ion search or replicate analysis
        ion_tolerance = args.precursor_ppm / 1e6 if args.mva else args.msn_ppm / 1e6
        ion_search_list = []
        replicate_search_list = defaultdict(list)
        scans_to_fetch = []

        # figure out the splines for mass accuracy correction
        calc_spline = (
            not mass_accuracy_correction and not raw_data_only and not args.neucode
        )
        spline_x = []
        spline_y = []
        spline = None

        scan_info_map = defaultdict(dict)

        for index, scan in enumerate(raw):
            if index % 100 == 0:
                sys.stderr.write(".")
            if scan is None:
                continue
            scan_id = scan.id
            msn_map.append((scan.ms_level if not args.mrm else scan.mass, scan_id))
            rt = scan.rt
            if scan.ms_level == msn_for_quant:
                msn_rt_map[scan_id] = int(scan.title) if args.mrm else rt
            scan_rt_map[scan_id] = rt
            scan_info_map[scan_id]["parent"] = scan.parent
            scan_info_map[scan_id]["msn"] = scan.ms_level
            scan_info_map[scan_id]["precursor"] = scan.mass
            scan_charge_map[scan_id] = scan.charge
            if msn_rt_window and not any(
                [(i[0] < float(rt) < i[1]) for i in msn_rt_window]
            ):
                continue
            if scan.parent:
                try:
                    scan_info_map[scan.parent]["children"].add(scan_id)
                except KeyError:
                    scan_info_map[scan.parent]["children"] = set([scan_id])
            if not scans_to_select or str(scan_id) in scans_to_select:
                if ion_search:
                    if scan.ms_level == msn_for_id:
                        scans_to_fetch.append(scan_id)
                elif all_msn:
                    # we are quantifying all msn spectra of a given type
                    if msn_for_id == scan.ms_level:
                        # find the closest scan to this, which will be the parent scan
                        spectra_to_quant = (
                            find_prior_scan(msn_map, scan_id, ms_level=msn_for_quant)
                            if msn_for_quant != msn_for_id
                            else scan_id
                        )
                        d = {
                            "quant_scan": {"id": spectra_to_quant},
                            "id_scan": {
                                "id": scan_id,
                                "rt": scan.rt,
                                "charge": scan.charge,
                                "mass": float(scan.mass),
                                "product_ion": float(scan.product_ion)
                                if args.mrm
                                else None,
                            },
                        }
                        ion_search_list.append((spectra_to_quant, d))
                elif args.mva:
                    if scan.ms_level == msn_for_quant:
                        scans_to_fetch.append(scan_id)
            if not raw_data_only and calc_spline:
                if hasattr(scan, "theor_mass"):
                    theor_mass = scan.getTheorMass()
                    observed_mass = scan.mass
                    mass_error = (theor_mass - observed_mass) / theor_mass * 1e6
                    spline_x.append(observed_mass)
                    spline_y.append(mass_error)
            del scan

        if calc_spline and len(spline_x):
            spline_df = pd.DataFrame(
                zip(spline_x, spline_y), columns=["Observed", "Error"]
            )
            spline_df = spline_df[
                (spline_df["Error"] < 25) & (spline_df["Error"] > -25)
            ].dropna()
            spline_df.sort("Observed", inplace=True)
            spline_df.drop_duplicates("Observed", inplace=True)
            if len(spline_df) > 10:
                spline = UnivariateSpline(
                    spline_df["Observed"].astype(float).values,
                    spline_df["Error"].astype(float).values,
                    s=1e6,
                )
            else:
                spline = None

        reader = Reader(
            reader_in,
            reader_outs,
            raw_file=filepath,
            spline=spline,
            rt_window=msn_rt_window,
        )
        reader.start()
        rep_map = defaultdict(set)
        if ion_search or args.mva:
            ions = (
                [
                    i["id_scan"].get("theor_mass", i["id_scan"]["mass"])
                    for i in raw_scans
                ]
                if args.mva
                else raw_scans["ions"]
            )
            rt_info = raw_scans.get("rt_info") if not args.mva else None
            last_scan_ions = defaultdict(set)
            for scan_id in scans_to_fetch:
                this_scan_ions = defaultdict(set)
                reader_in.put((0, scan_id, None, None))
                scan = reader_outs[0].get()
                if scan is None:
                    continue
                if args.msn_all_scans:
                    # if not args.require_all_ions:
                    #     ions = set([j for i in ions for j in i])
                    for ion_index, ion_set in enumerate(ions):
                        d = {
                            "quant_scan": {"id": scan_id, "scans": scans_to_fetch},
                            "id_scan": {
                                "id": scan_id,
                                "theor_mass": ion_set[0],
                                "rt": rt_info[ion_index] if rt_info else scan["rt"],
                                "charge": 1,
                                "mass": ion_set[0],
                                "ions_found": None,
                                "ion_set": ion_set,
                            },
                            "combine_xics": args.require_all_ions,
                        }
                        ion_search_list.append((scan_id, d))
                    break
                scan_mzs = scan["vals"]
                mz_vals = scan_mzs[scan_mzs[:, 1] > 0][:, 1]
                scan_mzs = scan_mzs[scan_mzs[:, 1] > 0][:, 0]
                if not np.any(scan_mzs):
                    continue
                mass, charge, rt, centroided = (
                    scan["mass"],
                    scan["charge"],
                    scan["rt"],
                    scan["centroid"],
                )
                ions_found = []
                added = set([])
                for ion_set in ions:
                    if isinstance(ion_set, float):
                        ion_set = [ion_set]
                    for ion_index, (ion, nearest_mz_index) in enumerate(
                        zip(
                            ion_set,
                            find_nearest_indices(
                                scan_mzs, np.array(ion_set, dtype=np.float)
                            ),
                        )
                    ):
                        nearest_mz = scan_mzs[nearest_mz_index]
                        found_ions = []
                        if peaks.get_ppm(ion, nearest_mz) < ion_tolerance:
                            if args.mva:
                                d = raw_scans[ion_index]
                                scan_rt = d["id_scan"]["rt"]
                                if (
                                    scan_rt - args.rt_window
                                    < rt
                                    < scan_rt + args.rt_window
                                ):
                                    found_ions.append(
                                        {
                                            "ion": ion,
                                            "nearest_mz": nearest_mz,
                                            "charge": d["id_scan"]["charge"],
                                            "scan_info": d,
                                            "ion_set": ion_set,
                                        }
                                    )
                            else:
                                ion_rt = rt_info[ion_index] if rt_info else None
                                if ion_rt is None or (
                                    rt - args.rt_window < ion_rt < rt + args.rt_window
                                ):
                                    found_ions.append(
                                        {
                                            "ion": ion,
                                            "nearest_mz": nearest_mz,
                                            "rt": ion_rt,
                                            "ion_set": ion_set,
                                        }
                                    )
                        elif args.require_all_ions:
                            # this means an ion in the set was not found, if we require all ions, continue to the next set
                            continue
                        # if we require all ions, only put in a single copy since we'll be finding the same ions in every
                        # scan anyways, otherwise put the entire set in since we may be missing ions
                        ions_found += (
                            found_ions[0] if args.require_all_ions else found_ions
                        )
                if ions_found:
                    # we have two options here. If we are quantifying a preceeding scan or the ion itself per scan
                    isotope_ppm = args.isotope_ppm / 1e6
                    if msn_for_quant == msn_for_id or args.mva:
                        for ion_dict in ions_found:
                            ion, nearest_mz = ion_dict["ion"], ion_dict["nearest_mz"]
                            ion_found = ",".join(map(str, ion_dict["ion_set"]))
                            spectra_to_quant = scan_id
                            # we are quantifying the ion itself
                            if charge == 0 and not args.mva:
                                # see if we can figure out the charge state
                                charge_states = []
                                for i in range(1, 5):
                                    charge_peaks_found = 0
                                    peak_height = 0
                                    for j in range(1, 3):
                                        next_peak = ion + CARBON_NEUTRON / float(
                                            i
                                        ) * float(j)
                                        closest_mz = peaks.find_nearest_index(
                                            scan_mzs, next_peak
                                        )
                                        if (
                                            peaks.get_ppm(
                                                next_peak, scan_mzs[closest_mz]
                                            )
                                            < isotope_ppm * 1.5
                                        ):
                                            charge_peaks_found += 1
                                            peak_height += mz_vals[closest_mz]
                                    charge_states.append(
                                        (charge_peaks_found, i, peak_height)
                                    )
                                charge_states = sorted(
                                    charge_states,
                                    key=operator.itemgetter(0, 2),
                                    reverse=True,
                                )
                                if int(ion_dict["charge"]) not in [
                                    i[0] for i in charge_states
                                ]:
                                    continue
                                elif args.mva:
                                    charge_to_use = ion_dict["charge"]
                                elif charge_states:
                                    if charge_states[0][1] == 1:
                                        charge_to_use = (
                                            charge_states[1][1]
                                            if charge_states[1][2] != 0
                                            else 1
                                        )
                                    else:
                                        charge_to_use = charge_states[0][1]
                                else:
                                    charge_to_use = 1
                            else:
                                charge_to_use = charge
                                if args.mva:
                                    rep_key = (
                                        ion_dict["scan_info"]["id_scan"]["rt"],
                                        ion_dict["scan_info"]["id_scan"]["mass"],
                                        charge_to_use,
                                    )
                                    rep_map[rep_key].add(rt)
                            this_scan_ions[ion].add(charge_to_use)
                            if charge_to_use in last_scan_ions[ion]:
                                continue
                            last_scan_ions[ion].add(charge_to_use)
                            if args.mva:
                                d = copy.deepcopy(ion_dict["scan_info"])
                                d["id_scan"]["id"] = scan_id
                                theo_mass = d["id_scan"].get("theor_mass")
                                if theo_mass:
                                    d["id_scan"]["mass"] = theo_mass
                                spectra_to_quant = find_prior_scan(
                                    msn_map, scan_id, ms_level=msn_for_quant
                                )
                                d["quant_scan"]["id"] = spectra_to_quant
                                d["replicate_scan_rt"] = rt
                            else:
                                d = {
                                    "quant_scan": {"id": scan_id},
                                    "id_scan": {
                                        "id": scan_id,
                                        "theor_mass": ion,
                                        "rt": rt if ion_rt is None else ion_rt,
                                        "charge": charge_to_use,
                                        "mass": float(nearest_mz),
                                        "ions_found": ion_found,
                                        "ion_set": ion_dict["ion_set"],
                                    },
                                    "combine_xics": True,
                                }
                            key = (scan_id, d["id_scan"]["theor_mass"], charge_to_use)
                            if key in added:
                                continue
                            added.add(key)
                            if args.mva:
                                replicate_search_list[
                                    (d["id_scan"]["rt"], ion_dict["ion"])
                                ].append((spectra_to_quant, d))
                            else:
                                ion_search_list.append((spectra_to_quant, d))
                    else:
                        # we are identifying the ion in a particular scan, and quantifying a preceeding scan
                        # find the closest scan to this, which will be the parent scan
                        spectra_to_quant = find_prior_scan(
                            msn_map, scan_id, ms_level=msn_for_quant
                        )
                        d = {
                            "quant_scan": {"id": spectra_to_quant},
                            "id_scan": {
                                "id": scan_id,
                                "rt": rt,
                                "charge": charge,
                                "mass": float(mass),
                                "ions_found": ";".join(
                                    map(
                                        lambda x: "{}({})".format(ion, nearest_mz),
                                        ions_found,
                                    )
                                ),
                            },
                        }
                        ion_search_list.append((spectra_to_quant, d))
                to_remove = []
                for ion, charges in last_scan_ions.items():
                    for charge in charges:
                        if charge not in this_scan_ions[ion]:
                            to_remove.append((ion, charge))
                for ion, charge in to_remove:
                    last_scan_ions[ion].discard(charge)
                del scan

        if args.mva:
            x = []
            y = []
            for i, v in rep_map.items():
                for j in v:
                    x.append(i[0])
                    y.append(j)
            from sklearn.linear_model import LinearRegression

            rep_mapper = LinearRegression()
            try:
                rep_mapper.fit(np.array(x).reshape(len(x), 1), y)
            except ValueError:
                rep_mapper = None

        if ion_search or all_msn:
            raw_scans = [
                i[1] for i in sorted(ion_search_list, key=operator.itemgetter(0))
            ]
        if args.mva:
            raw_scans = []
            for i in replicate_search_list:
                ion_rt, ion = i
                best_scan = sorted(
                    [
                        (
                            np.abs(
                                (
                                    rep_mapper.predict(
                                        np.array(j[1]["replicate_scan_rt"]).reshape(
                                            -1, 1
                                        )
                                    )
                                    if rep_mapper
                                    else j[1]["replicate_scan_rt"]
                                )
                                - ion_rt
                            ),
                            j[1],
                        )
                        for j in replicate_search_list[i]
                    ],
                    key=operator.itemgetter(0),
                )[0][1]
                raw_scans.append(best_scan)

        quant_msn_map = (
            [i for i in msn_map if i[0] == msn_for_quant] if not args.mrm else msn_map
        )
        manager = Manager()
        scan_mask = manager.dict()

        for i in range(threads):
            worker = Worker(
                queue=in_queue,
                results=result_queue,
                raw_name=filepath,
                mass_labels=mass_labels,
                debug=args.debug,
                html=html,
                mono=not args.spread,
                precursor_ppm=args.precursor_ppm,
                isotope_ppm=args.isotope_ppm,
                isotope_ppms=None,
                msn_rt_map=msn_rt_map,
                reporter_mode=reporter_mode,
                reader_in=reader_in,
                reader_out=reader_outs[i],
                thread=i,
                quant_method=quant_method,
                spline=spline,
                isotopologue_limit=isotopologue_limit,
                labels_needed=labels_needed,
                quant_msn_map=quant_msn_map,
                overlapping_mz=overlapping_mz,
                min_resolution=args.min_resolution,
                min_scans=args.min_scans,
                mrm_pair_info=mrm_pair_info,
                mrm=args.mrm,
                peak_cutoff=args.peak_cutoff,
                replicate=args.mva,
                ref_label=ref_label,
                max_peaks=args.max_peaks,
                parser_args=args,
                scans_to_skip=scan_mask,
            )
            workers.append(worker)
            worker.start()

        # TODO:
        # combine information from scans (where for instance, we have fragmented both the heavy/light
        # peptides -- we want to use those masses before calculating where it should be). This may not
        # be possible for all types of input though, figure this out.
        quant_map = {}
        scans_to_submit = []

        # this is to fix the header at the end to include peak information if we have multiple peaks
        most_peaks_found = 0
        mrm_added = set([])
        exclusion_masses = (
            mrm_pair_info.loc[
                :,
                [
                    i
                    for i in mrm_pair_info.columns
                    if i.lower() not in ("light", "retention time")
                ],
            ].values.flatten()
            if args.mrm
            else set([])
        )
        for scan_index, raw_scan_info in enumerate(raw_scans):
            target_scan = raw_scan_info["id_scan"]
            quant_scan = raw_scan_info["quant_scan"]
            scanId = target_scan["id"]
            scan_mass = target_scan.get("mass")
            if args.mrm:
                if scan_mass in mrm_added:
                    continue
                mrm_added.add(scan_mass)
                if scan_mass in exclusion_masses:
                    continue

            if quant_scan.get("id") is None:
                scan_to_quant = None
                if msn_for_quant > msn_for_id:
                    # Cases like MS3 hit this, go down until we hit the correct MS level and add
                    # those scans to be quantified
                    children = scan_info_map[scanId].get("children", [scanId])
                    quant_scan["scans"] = []
                    while children:
                        child = children.pop()
                        child_scan_info = scan_info_map[child]
                        scan_to_quant_ms = child_scan_info["msn"]
                        if scan_to_quant_ms < msn_for_quant:
                            children += child_scan_info.get("children", [])
                        if scan_to_quant_ms == msn_for_quant:
                            if scan_to_quant is None:
                                scan_to_quant = child
                            quant_scan["scans"].append(child)
                else:
                    # we will hit this in a normal proteomic run
                    # figure out the ms-1 from the ms level we are at
                    scan_info = scan_info_map[scanId]
                    current_scan = scan_info["parent"]
                    try:
                        while current_scan:
                            scan_info = scan_info_map[scanId]
                            current_scan = scan_info["parent"]
                            scan_to_quant_ms = scan_info[scan_to_quant]["msn"]
                            if scan_to_quant_ms == msn_for_quant:
                                scan_to_quant = current_scan
                    except KeyError:
                        scan_to_quant = None
                if scan_to_quant is not None:
                    msn_to_quant = scan_to_quant
                else:
                    msn_to_quant = find_prior_scan(
                        msn_map, scanId, ms_level=msn_for_quant
                    )
                quant_scan["id"] = msn_to_quant

            rt = target_scan.get("rt", scan_rt_map.get(scanId))
            if rt is None:
                rt = float(msn_rt_map[msn_to_quant])
                target_scan["rt"] = rt

            if args.mva and rep_mapper is not None:
                target_scan["rt"] = rep_mapper.predict(
                    np.array(float(target_scan["rt"])).reshape(-1, 1)
                )[0]

            mods = target_scan.get("modifications")
            charge = target_scan.get("charge")
            if charge is None or charge == 0:
                charge = int(scan_charge_map.get(scanId, 0))
            if charge == 0:
                continue
            charge = int(charge)

            if msn_for_quant != 1:
                lowest_label = min([j for i, v in mass_labels.items() for j in v])
                target_scan["theor_mass"] = lowest_label
                target_scan["precursor"] = lowest_label
            else:
                mass_shift = 0

                if mods is not None:
                    shift = 0
                    for mod in filter(lambda x: x, mods.split("|")):
                        aa, pos, mass, _ = mod.split(",", 3)
                        mass = get_formatted_mass(mass)
                        if aa in silac_shifts.get(mass, {}):
                            shift += mass
                    mass_shift += float(shift) / float(charge)
                else:
                    # assume we are the light version, include all the labels we are looking for here
                    pass

                target_scan["theor_mass"] = (
                    target_scan.get("theor_mass", target_scan.get("mass")) - mass_shift
                )
                target_scan["precursor"] = (
                    target_scan["mass"] - mass_shift
                    if not args.neucode
                    else target_scan["theor_mass"]
                )
            # key is filename, peptide, charge, target scan id, modifications
            key = (
                filename,
                target_scan.get("peptide", ""),
                target_scan.get("charge"),
                target_scan.get("id"),
                target_scan.get("modifications"),
            )
            if resume:
                if tuple(map(str, key)) in skip_map:
                    completed += 1
                    continue
            params = {"scan_info": raw_scan_info}
            scans_to_submit.append((target_scan["rt"], params))

        # sort by RT so we can minimize our memory footprint by throwing away scans we no longer need
        scans_to_submit.sort(key=operator.itemgetter(0))
        if ion_search or all_msn or args.mva:
            scan_count = len(scans_to_submit)
        for i in scans_to_submit:
            in_queue.put(i[1])

        sys.stderr.write("{0} processed and placed into queue.\n".format(filename))

        # kill the workers
        [in_queue.put(None) for i in range(threads)]
        RESULT_DICT = {i[0]: "NA" for i in RESULT_ORDER}
        scans_to_export = set([])
        export_mapping = defaultdict(set)
        key_map = list(msn_rt_map.keys())
        rt_scan_map = pd.Series(key_map, index=[msn_rt_map[i] for i in key_map])
        rt_scan_map.sort_index(inplace=True)

        result = None
        while workers or result is not None:
            try:
                result = result_queue.get(timeout=0.1)
            except Empty:
                # kill expired workers
                result = None
                to_del = []
                for i, v in enumerate(workers):
                    if not v.is_alive():
                        v.terminate()
                        exit_code = v.exitcode
                        if exit_code in CRASH_SIGNALS:
                            print(
                                "thread has been killed using params {}".format(
                                    v.params
                                )
                            )
                        to_del.append(
                            {
                                "worker_index": i,
                                "thread_id": v.thread,
                                "exitcode": exit_code,
                            }
                        )
                workers_to_add = []
                for worker_dict in sorted(
                    to_del, key=operator.itemgetter("worker_index"), reverse=True
                ):
                    worker_index = worker_dict["worker_index"]
                    if worker_dict["exitcode"] in CRASH_SIGNALS:
                        thread_index = worker_dict["thread_id"]
                        worker = Worker(
                            queue=in_queue,
                            results=result_queue,
                            raw_name=filepath,
                            mass_labels=mass_labels,
                            debug=args.debug,
                            html=html,
                            mono=not args.spread,
                            precursor_ppm=args.precursor_ppm,
                            isotope_ppm=args.isotope_ppm,
                            isotope_ppms=None,
                            msn_rt_map=msn_rt_map,
                            reporter_mode=reporter_mode,
                            reader_in=reader_in,
                            reader_out=reader_outs[thread_index],
                            thread=thread_index,
                            quant_method=quant_method,
                            spline=spline,
                            isotopologue_limit=isotopologue_limit,
                            labels_needed=labels_needed,
                            quant_msn_map=[i for i in msn_map if i[0] == msn_for_quant]
                            if not args.mrm
                            else msn_map,
                            overlapping_mz=overlapping_mz,
                            min_resolution=args.min_resolution,
                            min_scans=args.min_scans,
                            mrm_pair_info=mrm_pair_info,
                            mrm=args.mrm,
                            peak_cutoff=args.peak_cutoff,
                            replicate=args.mva,
                            ref_label=ref_label,
                            max_peaks=args.max_peaks,
                            parser_args=args,
                        )
                        workers_to_add.append(worker)
                        worker.start()
                    del workers[worker_index]
                workers += workers_to_add
            if result is not None:
                completed += 1
                if completed % 10 == 0:
                    sys.stderr.write(
                        "\r{0:2.2f}% Completed".format(completed / scan_count * 100)
                    )
                    sys.stderr.flush()
                res_dict = copy.deepcopy(RESULT_DICT)
                for i in RESULT_ORDER:
                    res_dict[i[0]] = result.get(i[0], "NA")
                peak_report = []
                for label_name in labels:
                    peaks_found = result.get("{}_peaks".format(label_name), {})
                    if args.export_mzml:
                        from . import PER_FILE, PER_ID, PER_PEAK

                        scans = get_scans_under_peaks(rt_scan_map, peaks_found)
                        flattened_scans = set(
                            [
                                l
                                for i, v in scans.items()
                                for j, k in v.items()
                                for l in k
                            ]
                        )
                        scans_to_export |= flattened_scans
                        if args.export_mode == PER_FILE:
                            export_mapping[
                                "{}_{}.mzML".format(out_path, filename)
                            ] |= flattened_scans
                    if len(peaks_found) > most_peaks_found:
                        most_peaks_found = len(peaks_found)

                    xic_peak_summary = OrderedDict()
                    for isotope_index, isotope_peaks in peaks_found.items():
                        if args.export_mzml and args.export_mode == PER_ID:
                            export_mapping[
                                "{out}_{raw}_{ms1}_{precursor}.mzML".format(
                                    **{
                                        "out": out_path,
                                        "raw": filename,
                                        "precursor": result.get(
                                            "{}_precursor".format(label_name)
                                        ),
                                        "ms1": result.get("ms1"),
                                    }
                                )
                            ] |= set(
                                [l for i, v in scans[isotope_index].items() for l in v]
                            )
                        for xic_peak_index, xic_peak_info in isotope_peaks.items():
                            if args.export_mzml and args.export_mode == PER_PEAK:
                                export_mapping[
                                    "{out}_{raw}_{ms1}_{precursor}_{isotope}_{peak}.mzML".format(
                                        **{
                                            "out": out_path,
                                            "raw": filename,
                                            "peak": xic_peak_index,
                                            "isotope": isotope_index,
                                            "precursor": result.get(
                                                "{}_precursor".format(label_name)
                                            ),
                                            "ms1": result.get("ms1"),
                                        }
                                    )
                                ] |= scans[isotope_index][xic_peak_index]
                            if args.peaks_n == 1:
                                # TODO: fix this terrible loop
                                for i, v in enumerate(RESULT_ORDER):
                                    for j in xic_peak_info:
                                        if "{}_{}".format(label_name, j) == v[0]:
                                            try:
                                                xic_peak_summary[i].append(
                                                    xic_peak_info[j]
                                                )
                                            except KeyError:
                                                xic_peak_summary[i] = [xic_peak_info[j]]
                            else:
                                peak_report.append(
                                    list(
                                        map(
                                            str,
                                            (
                                                xic_peak_info.get(i[0], "NA")
                                                for i in PEAK_REPORTING
                                            ),
                                        )
                                    )
                                )

                    if args.peaks_n == 1:
                        for index, values in xic_peak_summary.items():
                            result_info = RESULT_ORDER[index]
                            if callable(result_info[-1]):
                                res_dict[result_info[0]] = result_info[-1](values)
                            else:
                                res_dict[result_info[0]] = values[0]
                res_dict["filename"] = filename
                # first entry of peak report is label, sort alphabetically
                peak_report.sort(key=operator.itemgetter(0))
                res_dict["peak_report"] = peak_report
                # This is the tsv output we provide
                res_list = (
                    [filename]
                    + [res_dict.get(i[0], "NA") for i in RESULT_ORDER]
                    + ["\t".join(i) for i in peak_report]
                )
                res = "{0}\n".format("\t".join(map(str, res_list)))
                out.write(res)
                out.flush()

                temp_file.write(
                    pd.io.json.dumps(
                        {"res_dict": res_dict, "html": result.get("html", {})}
                    )
                )
                temp_file.write("\n")
                temp_file.flush()

        if scans_to_export:
            raw_file = GuessIterator(open(filepath, "rb"), full=True, store=False)
            for export_filename, scans in export_mapping.items():
                with open(export_filename, "wb") as o:
                    raw_file.parser.writeScans(handle=o, scans=sorted(scans))

        reader_in.put(None)

        del msn_map
        del scan_rt_map
        del raw

    out.flush()
    out.close()
    # fix the header if we need to. We reopen the file because Windows doesn't like it when we read on a file with 'w+'
    if args.peaks_n != 1 and out != sys.stdout:
        tmp_file = "{}_ot".format(out_path)
        with open(out.name, "r") as out:
            with open(tmp_file, "w") as o:
                new_header = out.readline().strip().split("\t")
                for peak_num in range(most_peaks_found):
                    new_header.extend(
                        ["Peak {} {}".format(peak_num, i[1]) for i in PEAK_REPORTING]
                    )
                o.write("{}\n".format("\t".join(new_header)))
                for row in out:
                    o.write(row)
        import shutil

        shutil.move(tmp_file, out_path)

    temp_file.close()
    df_data = []
    html_data = []
    peak_data = []

    for j in open(temp_file.name, "r"):
        result = pd.io.json.loads(j if isinstance(j, str) else j.decode("utf-8"))
        res_dict = result["res_dict"]
        res_list = [res_dict["filename"]] + [
            res_dict.get(i[0], "NA") for i in RESULT_ORDER
        ]
        peak_data.append(res_dict["peak_report"])
        df_data.append(res_list)
        html_data.append(result["html"])
    data = pd.DataFrame.from_records(
        df_data, columns=[i for i in headers if i != "Confidence"]
    )
    header_mapping = []
    order_names = [i[1] for i in RESULT_ORDER]
    for i in data.columns:
        try:
            header_mapping.append(RESULT_ORDER[order_names.index(i)][0])
        except ValueError:
            header_mapping.append(False)
    if calc_stats:
        perform_ml(data, mass_labels)
        data.to_csv("{}_stats".format(out_path), sep=str("\t"), index=None)

    if html:
        html_map = []
        peak_map = {"header": [i[1] for i in PEAK_REPORTING], "data": []}
        for index, (row_index, row) in enumerate(data.iterrows()):
            res = "\t".join(row.astype(str))
            to_write, html_info = table_rows(
                [
                    {
                        "table": res.strip(),
                        "html": html_data[index],
                        "keys": header_mapping,
                    }
                ]
            )
            html_map.append(html_info)
            peak_map["data"].append(peak_data[index])
            html_out.write(to_write)
            html_out.flush()

        template = []
        append = False
        for i in open(pyquant_html_file, "r"):
            if "HTML BREAK" in i:
                append = True
            elif append:
                template.append(i)
        html_template = Template("".join(template))
        peak_map = pd.io.json.dumps(peak_map)
        html_map = pd.io.json.dumps(html_map)
        html_out.write(
            html_template.safe_substitute(
                {
                    "peak_output": base64.b64encode(
                        gzip.zlib.compress(bytes(peak_map, "utf-8"), 9,)
                    ).decode("utf-8"),
                    "html_output": base64.b64encode(
                        gzip.zlib.compress(bytes(html_map, "utf-8"), 9,)
                    ).decode("utf-8"),
                }
            )
        )

    os.remove(temp_file.name)



import pickle
import os
from functools import wraps


def timer(func):
    @wraps(func)
    def inner(*args, **kwargs):
        import time

        start = time.time()
        val = func(*args, **kwargs)
        elapsed = time.time() - start
        print("{} evaluated in {}".format(func.__name__, elapsed))
        return val

    return inner


def update_pickle(key, data, overwrite=False):
    base_dir = os.path.split(__file__)[0]
    pickle_file = os.path.join(base_dir, "data", "peak_data.pickle")
    with open(pickle_file, "rb") as f:
        p = pickle.load(f)
    if key in p and not overwrite:
        raise Exception(
            "{} already exists. Pass overwrite=True to overwrite existing data.".format(
                key
            )
        )
    p[key] = data
    with open(pickle_file, "wb") as o:
        pickle.dump(p, o)



import os
import subprocess
import unittest

import numpy as np
import pandas as pd

from . import config
from .mixins import FileMixins
from .utils import timer


peps_scans = (
    # possible incomplete envelope identification but good fits
    "GQGYLFELR 5059",
    "IFLDASSEER 3147",
    "ATIGEVGNAEHMLR 2607",
    "YMANAMGPEGVR 1975",
    "VPDSQVLADLDHVASWASR 7203",
    "VPGTIGFATVR 3830",
    "VVANFLSSVGVDR 5699",
    "IHVAVAQEVPGTGVDTPEDLER 4157",
    "INAMLQDYELQR 4442",
    "AFIGIDGWQPETGFTGR 6770",
    "GMNVVFELR 5091",
    "ILHIQQQLAGEQVALSDEVNQSEQTTNFHNR 5404",
    "EQGYALDSEENEQGVR 2536",
    "VGSDNLLMINAHIAHDCTVGNR 4150",
    "FDGFVHSIGFAPGDQLDGDYVNAVTR 6357",
    "GIWNHGSPLFMEIEPR 6173",
    # bad data
    "VGYIELDLNSGK 6246",
    "FVESVDVAVNLGIDAR 9184",
    # bad fit
    "TIPSVLTALFCAR 8606",
    "VANLEAQLAEAQTR 3878",
    # uncertain
    "ITGIDSSPAMIAEAR 4149",
    # good fit
    "TVYSTENPDLLVLEFR 7102",
    "VMMIDEPAILDQAIAR 6949",
    "IIAIDNSPAMIER 4301",
    "IIAIDNSPAMIER 4307",
    "DAGNIIIDDDDISLLPLHAR 6921",
    "NASASAPTALPLR 3089",
    "SLEHEVTLVDDTLAR 4955",
    "MTELPIDENTPR 3494",
    "LAQALANPLFPALDSALR 7905",
    "SDWNPSLYLHFSAER 7079",
    # could be better at complete fit
    "GIWNHGSPLFMEIEPR 6175",
    "VANLEAQLAEAQTR 3835",
    "HVAIIMDGNGR 2069",
    "GAGEIVLNMMNQDGVR 6081",
    "ITGIDSSPAMIAEAR 4159",
    # could be better/maybe incomplete envelope
    "FDGFVHSIGFAPGDQLDGDYVNAVTR 6747",
    "DAGNIIIDDDDISLLPLHAR 6930",
    "MQQLQNIIETAFER 7023",
    "DQQIPLLISGGIGHSTTFLYSAIAQHPHYNTIR 7545",
)


class EColiManualTest(FileMixins, unittest.TestCase):
    def setUp(self):
        super(EColiManualTest, self).setUp()
        self.output = os.path.join(self.out_dir, "pqtest_manual")
        self.output_stats = os.path.join(self.out_dir, "pqtest_stats")

    @timer
    def test_pyquant_manual(self):
        com = [
            self.executable,
            "--search-file",
            self.ecoli_search_file,
            "--scan-file",
            self.ecoli_mzml,
            "--html",
            "-p",
            str(config.CORES),
            "-o",
            self.output,
            "--xic-window-size",
            "12",
        ]
        peptides, scans = zip(*map(lambda x: x.split(" "), peps_scans))
        com.extend(["--peptide"])
        com.extend(peptides)
        com.extend(["--scan"])
        com.extend(scans)
        subprocess.call(com)
        pyquant = pd.read_table(self.output)
        label = "Medium"
        pq_sel = "{}/Light".format(label)
        pyquant.loc[
            (
                (pyquant["Peptide"].str.upper().str.count("R") == 1)
                & (pyquant["Peptide"].str.upper().str.count("K") == 0)
            ),
            "Class",
        ] = "R"
        pyquant.loc[
            (
                (pyquant["Peptide"].str.upper().str.count("K") == 1)
                & (pyquant["Peptide"].str.upper().str.count("R") == 0)
            ),
            "Class",
        ] = "K"
        pyquant[pq_sel] = np.log2(pyquant[pq_sel] + 0.000001)
        label = "Heavy"
        pq_sel = "{}/Light".format(label)
        pyquant[pq_sel] = np.log2(pyquant[pq_sel] + 0.000001)


if __name__ == "__main__":
    unittest.main()



import time
import unittest
from multiprocessing import Queue

from pyquant.reader import Reader

from . import mixins


class TestReader(mixins.FileMixins, unittest.TestCase):
    def setUp(self):
        super(TestReader, self).setUp()
        self.input = Queue()
        self.output = {0: Queue()}
        self.reader = Reader(self.input, self.output, raw_file=self.small_mzml)
        self.addCleanup(lambda: self.input.put(None))
        self.reader.start()
        self.queue = self.output[0]

    def test_fetches_scan(self):
        self.input.put((0, "S19", 0, 10000))
        # Test we get the scan
        scan = self.queue.get()
        self.assertEqual(scan["title"], "S19")

    def test_fetches_subset(self):
        # test we get a subset
        self.input.put((0, "S19", 0, 10000))
        scan = self.queue.get()
        mz_min, mz_max = scan["vals"][2, 0], scan["vals"][5, 0]
        self.input.put((0, "S19", mz_min, mz_max))
        scan = self.queue.get()
        self.assertTrue(scan["vals"][0, 0] >= mz_min)
        self.assertTrue(scan["vals"][-1, 0] <= mz_max)

    def test_exits_on_sentinel(self):
        # test we exit on a sentinel
        self.input.put(None)
        # wait for the process to end
        time.sleep(0.5)
        self.assertFalse(self.reader.is_alive())

    def test_refetches_deleted_scans(self):
        # test that we can get scans that have been deleted
        self.reader.timeout_minutes = 0.01
        self.input.put((0, "S19", 0, 10000))
        self.input.put((0, "S20", 0, 10000))
        scan = self.queue.get()
        scan = self.queue.get()
        time.sleep(1)
        # accessing 81 should cause 80 to be deleted
        self.input.put((0, "S20", 0, 10000))
        scan = self.queue.get()
        # make sure we can get it back
        self.input.put((0, "S19", 0, 10000))
        scan = self.queue.get()
        self.assertEqual(scan["title"], "S19")


if __name__ == "__main__":
    unittest.main()



__author__ = "chris"
import pickle
import os
import unittest

import numpy as np
import pandas as pd
from nose.plugins.attrib import attr

from pyquant import utils
from .mixins import GaussianMixin


class UtilsTests(GaussianMixin, unittest.TestCase):
    def setUp(self):
        super(UtilsTests, self).setUp()
        self.base_dir = os.path.split(os.path.abspath(__file__))[0]
        self.data_dir = os.path.join(self.base_dir, "data")

    def test_savgol_smooth(self):
        y = np.array([0, 1, 1.5, 0])
        new_y = utils.savgol_smooth(y)
        self.assertAlmostEqual(new_y[1], 1.0)
        self.assertAlmostEqual(new_y[2], 1.5)

    def test_select_window(self):
        x = list(range(10))
        selection = utils.select_window(x, 0, 3)
        self.assertListEqual(selection, [0, 1, 2, 3])
        selection = utils.select_window(x, 3, 3)
        self.assertListEqual(selection, [0, 1, 2, 3, 4, 5, 6])
        selection = utils.select_window(x, 8, 3)
        self.assertListEqual(selection, [5, 6, 7, 8, 9])
        selection = utils.select_window(x, 8, 20)
        self.assertListEqual(selection, x)
        # Make sure we don't break if a float is passed
        selection = utils.select_window(x, 8, 20.0)
        self.assertListEqual(selection, x)

    def test_divide_peaks(self):
        chunks = utils.divide_peaks(self.one_gauss)
        two_gauss_chunks = utils.divide_peaks(self.two_gauss, chunk_factor=1.0)
        self.assertEqual(len(chunks), 0)
        self.assertEqual(len(two_gauss_chunks), 1)
        self.assertEqual(two_gauss_chunks[0], 65)

    def test_calculate_theoretical_distribution(self):
        peptide = "PEPTIDE"
        pep_comp = utils.calculate_theoretical_distribution(peptide=peptide)
        ele_comp = utils.calculate_theoretical_distribution(
            elemental_composition={"C": 7}
        )
        np.testing.assert_almost_equal(
            pep_comp.values.tolist(),
            [
                0.6411550319843632,
                0.2662471681269686,
                0.07401847648709056,
                0.015434213671511215,
                0.002681646815294711,
            ],
        )
        np.testing.assert_almost_equal(
            ele_comp.values.tolist(),
            [0.9254949240653104, 0.07205572209608584, 0.002404285974894674],
        )

    @attr("linux")
    def test_ml(self):
        data = os.path.join(self.data_dir, "ml_data.tsv")
        dat = pd.read_table(data)
        labels = ["Heavy", "Medium", "Light"]
        utils.perform_ml(dat, {i: [] for i in labels})
        for label1 in labels:
            for label2 in labels:
                if label1 == label2:
                    continue
                col = "{}/{} Confidence".format(label1, label2)
                self.assertNotEqual(sum(pd.isnull(dat["Heavy/Light Confidence"])), 0)

    def test_merge_peaks(self):
        peaks = {
            1: {"minima": [0, 1, 2, 4, 5], "peaks": [3]},
            2: {"minima": [0, 1, 2, 4, 5], "peaks": [3]},
            7: {"minima": [0, 1, 2, 4, 5], "peaks": [3]},
        }
        merged = utils.merge_peaks(peaks)
        self.assertDictEqual(merged, {7: {"minima": [0, 1, 2, 4, 5], "peaks": [3]}})

        peaks = {1: {"minima": [0, 1, 2, 4, 5], "peaks": [3]}}
        merged = utils.merge_peaks(peaks)
        self.assertDictEqual(merged, {1: {"minima": [0, 1, 2, 4, 5], "peaks": [3]}})

        peaks = {
            1: {"minima": [0, 5], "peaks": [3, 7, 8]},
            2: {"minima": [0, 5], "peaks": [3, 7]},
            7: {"minima": [0, 5], "peaks": [3, 7]},
        }
        merged = utils.merge_peaks(peaks)
        self.assertDictEqual(
            merged,
            {
                1: {"minima": [0, 5], "peaks": [3, 7, 8]},
                7: {"minima": [0, 5], "peaks": [3, 7]},
            },
        )

    def test_get_cross_points(self):
        y = [1, 1, 1, 1, 1, 1, 1]
        self.assertListEqual(utils.get_cross_points(y), [])

        y = [1, 1, 1, 1, 1, -1, 1]
        self.assertListEqual(utils.get_cross_points(y, pad=False), [4, 5])
        self.assertListEqual(utils.get_cross_points(y, pad=True), [0, 4, 5, 6])

        y = [1, -1, 1]
        self.assertListEqual(utils.get_cross_points(y, pad=False), [0, 1])
        self.assertListEqual(utils.get_cross_points(y, pad=True), [0, 1, 2])

        y = [1, 1, 1, 1, -1, -1, -1, -1]
        self.assertListEqual(utils.get_cross_points(y, pad=True), [0, 3, 7])

    def test_find_peaks_derivative(self):
        with open(os.path.join(self.data_dir, "peak_data.pickle"), "rb") as peak_file:
            data = pickle.load(peak_file, encoding="latin1")

        x, y = data["large_range"]
        peaks = utils.find_peaks_derivative(x, y, smooth=False)
        peaks = next(iter(peaks.values()))

        np.testing.assert_array_equal(
            peaks["peaks"],
            np.array(
                [
                    357,
                    378,
                    432,
                    1668,
                    1755,
                    1811,
                    1835,
                    1912,
                    2009,
                    2399,
                    2577,
                    2952,
                    3171,
                ]
            ),
        )
        np.testing.assert_array_equal(
            peaks["minima"],
            np.array(
                [
                    337,
                    366,
                    366,
                    423,
                    423,
                    667,
                    1654,
                    1678,
                    1732,
                    1774,
                    1798,
                    1825,
                    1825,
                    1844,
                    1901,
                    1930,
                    1992,
                    2024,
                    2369,
                    2409,
                    2543,
                    2608,
                    2905,
                    2996,
                    3155,
                    3187,
                ]
            ),
        )

    def test_interpolate_data(self):
        y = [
            0,
            0,
            0,
            6311371.403331924,
            24368020.237973947,
            33309587.186450623,
            0,
            0,
            22678022.890094325,
            12544950.520046625,
            9621327.844190728,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
        ]
        x = [
            23.6401,
            23.6813,
            23.7225,
            23.7650,
            23.8063,
            23.8484,
            23.8905,
            23.9343,
            23.9759,
            24.0183,
            24.0602,
            24.1009,
            24.1440,
            24.1877,
            24.2282,
            24.2738,
            24.3195,
            24.3662,
            24.4168,
            24.4607,
            24.5016,
            24.5417,
            24.5812,
            24.6235,
            24.7088,
            24.7942,
        ]
        interp_y = utils.interpolate_data(x, y, gap_limit=2)
        self.assertNotEqual(interp_y[6], 0)
        self.assertNotEqual(interp_y[7], 0)
        self.assertCountEqual(interp_y[:3], [0, 0, 0])
        interp_y = utils.interpolate_data(x, y, gap_limit=1)
        self.assertCountEqual(interp_y[6:8], [0, 0])

    def test_merge_close_peaks(self):
        ty = np.array([0, 1, 2, 1, 0, 1, 2, 3, 2, 1, 0, 1, 3, 3])
        merged = utils.merge_close_peaks(np.array([7, 12]), ty, distance=6)
        np.testing.assert_array_equal(merged, np.array([7, 12]))

        merged = utils.merge_close_peaks(np.array([2, 7, 12]), ty, distance=5)
        np.testing.assert_array_equal(merged, np.array([2, 7, 12]))

        ty = np.array([0, 1, 2, 1, 0, 1, 2, 4, 2, 1, 0, 1, 3, 3])
        merged = utils.merge_close_peaks(np.array([2, 7, 12]), ty, distance=6)
        np.testing.assert_array_equal(merged, np.array([7]))

        merged = utils.merge_close_peaks(np.array([]), ty, distance=6)
        np.testing.assert_array_equal(merged, np.array([]))

    def test_rel_extrema(self):
        from scipy.signal import argrelextrema

        y = np.array(
            [
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                720001.9375,
                3056311.5,
                5417428.5,
                5494928.0,
                2185260.5,
                4099960.25,
                6359202.5,
                7897908.5,
                5956936.5,
                3124577.75,
                533339.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                554049.75,
                2336927.0,
                4342714.5,
                4671674.0,
                3267909.5,
                2324766.75,
                3356381.25,
                5258928.5,
                6305622.5,
                4235590.0,
                2183502.75,
                306638.96875,
                0.0,
                0.0,
                0.0,
                0.0,
            ]
        )
        self.assertTrue(
            np.array_equal(
                argrelextrema(y, np.greater), utils.argrelextrema(y, np.greater)
            )
        )

    def test_get_formatted_mass(self):
        self.assertEqual(
            utils.get_formatted_mass("0.123"), utils.get_formatted_mass(0.123)
        )
        self.assertEqual(
            utils.get_formatted_mass("0.12300"), utils.get_formatted_mass(0.123)
        )
        self.assertEqual(
            utils.get_formatted_mass("123.12300"), utils.get_formatted_mass(123.123)
        )
        self.assertEqual(
            utils.get_formatted_mass("123.12300"), utils.get_formatted_mass(123.1230000)
        )

    def test_get_scan_resolution(self):
        with open(os.path.join(self.data_dir, "peak_data.pickle"), "rb") as peak_file:
            data = pickle.load(peak_file, encoding="latin1")

        x, y = data["low_res_scan"]
        scan = pd.Series(y, index=x)
        resolution = utils.get_scan_resolution(scan)
        self.assertAlmostEqual(int(resolution), 30459)

        x = np.array(
            [
                805.47264226,
                805.47265226,
                805.47495304,
                805.47739824,
                805.47984345,
                805.49206968,
                805.49451496,
                805.49696025,
                805.49940556,
                805.50185087,
                805.50429619,
                805.50674153,
                805.50918687,
                805.51163223,
                805.51407748,
                805.51652286,
                805.51896825,
                805.52141365,
                805.7635636,
                805.76601011,
                805.76845664,
                805.77090318,
                805.77334973,
                805.77579629,
                805.77824286,
                805.78068944,
                805.78313604,
                805.78558264,
                805.79153526,
                805.79398188,
                805.79642852,
                805.79887517,
                805.80132183,
                805.8037685,
                805.80621519,
                805.80622519,
                805.80623519,
                805.80661184,
                805.80905857,
                806.09295462,
                806.09540266,
                806.0978507,
                806.10029875,
                806.10274682,
                806.1051949,
                806.10764298,
                806.11009108,
                806.11253919,
                806.11498731,
                806.11743545,
                806.12539827,
                806.12784643,
                806.13029459,
                806.13274277,
                806.13519096,
                806.13763916,
                806.14008737,
                806.14009737,
                806.14206697,
                806.14451522,
            ]
        )
        y = np.array(
            [
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                720001.9375,
                3056311.5,
                5417428.5,
                5494928.0,
                2185260.5,
                4099960.25,
                6359202.5,
                7897908.5,
                5956936.5,
                3124577.75,
                533339.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                554049.75,
                2336927.0,
                4342714.5,
                4671674.0,
                3267909.5,
                2324766.75,
                3356381.25,
                5258928.5,
                6305622.5,
                4235590.0,
                2183502.75,
                306638.96875,
                0.0,
                0.0,
                0.0,
                0.0,
            ]
        )

        scan = pd.Series(y, index=x)
        resolution = utils.get_scan_resolution(scan)
        self.assertAlmostEqual(int(resolution), 90781)

    def test_subtract_baseline(self):
        # This test data is a simple sin function with a parabola function applied to the ends,
        # giving a smile shape.

        x = np.linspace(-50, 50)
        noise = 0.00125 * x ** 2
        c1 = np.sin(x)
        y = noise + c1

        baselined = utils.subtract_baseline(y)

        # Assert we're about zero and the noise (the parabola) has been removed
        self.assertTrue(sum(baselined - c1) < 1e-10)


class TestGetScansUnderPeak(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        super(TestGetScansUnderPeak, cls).setUpClass()
        cls.rt_scan_map = pd.Series({1 + i / 10.0: i for i in range(30)})
        cls.found_peaks = {
            0: {
                0: {"peak_mean": 2, "std": 0.3, "std2": 0.2,},
                1: {"peak_mean": 1.3, "std": 0.2, "std2": 0.45,},
            }
        }

    def test_gets_peaks_under_scan(self):
        scans = utils.get_scans_under_peaks(self.rt_scan_map, self.found_peaks)
        self.assertEqual(
            scans,
            {
                0: {
                    0: {4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14},
                    1: {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12},
                }
            },
        )


if __name__ == "__main__":
    unittest.main()



import pickle
import os

import numpy as np

BASE_DIR = os.path.split(os.path.abspath(__file__))[0]
DATA_DIR = os.path.join(BASE_DIR, "data")


class FileMixins(object):
    def setUp(self):
        super(FileMixins, self).setUp()
        self.base_dir = BASE_DIR
        self.data_dir = DATA_DIR
        self.executable = os.path.abspath(
            os.path.join(self.base_dir, "..", "scripts", "pyQuant")
        )
        self.ecoli_search_file = os.path.join(
            self.data_dir, "Chris_Ecoli_1-2-4-(01).msf"
        )
        self.ecoli_mzml = os.path.join(self.data_dir, "Chris_Ecoli_1-2-4.mzML")
        self.itraq_mzml = os.path.join(self.data_dir, "iTRAQ_Data.mzML")
        self.small_mzml = os.path.join(self.data_dir, "tiny.msdata.mzML0.99.10.mzML")
        self.out_dir = "pq_tests"
        try:
            os.mkdir(self.out_dir)
        except OSError:
            pass

    # def tearDown(self):
    #     os.remove(self.output)
    #     os.remove(self.output_stats)
    #     os.rmdir(self.out_dir)


class GaussianMixin(object):
    def setUp(self):
        self.amp, self.mu, self.std, self.mu2 = 1.0, 0.0, 1.0, 3.0
        self.one_gauss_params = np.array([self.amp, self.mu, self.std], dtype=np.float)
        self.two_gauss_params = np.array(
            [self.amp, self.mu, self.std, self.amp, self.mu2, self.std], dtype=np.float
        )
        with open(os.path.join(DATA_DIR, "peak_data.pickle"), "rb") as peak_file:
            self.peak_data = pickle.load(peak_file, encoding="latin1")
        self.x = self.peak_data["one_gauss"][0]
        self.one_gauss = self.peak_data["one_gauss"][1]
        self.two_gauss = self.peak_data["two_gauss"][1]
        self.noisy_two_gauss = self.peak_data["noisy_two_gauss"][1]



import copy
import os
import sys
import warnings
from collections import Counter
from functools import partial
from itertools import combinations
from operator import itemgetter

import six
import numpy as np
import pandas as pd
from scipy.signal import savgol_filter
from scipy.special import comb

from . import PEAK_FINDING_DERIVATIVE, PEAK_FINDING_REL_MAX
from .logger import logger


ERRORS = "error" if os.environ.get("PYQUANT_DEV", False) == "True" else "ignore"


def merge_list(starting_list):
    final_list = []
    for i, v in enumerate(starting_list[:-1]):
        if set(v) & set(starting_list[i + 1]):
            starting_list[i + 1].extend(list(set(v) - set(starting_list[i + 1])))
        else:
            final_list.append(v)
    final_list.append(starting_list[-1])
    return final_list


def findValleys(y, srt):
    peak = y.iloc[srt]
    for left in range(srt - 1, -1, -1):
        val = y.iloc[left]
        if val == 0 or val > peak:
            break
    right = len(y)
    for right in range(srt + 1, len(y)):
        val = y.iloc[right]
        if val == 0 or val > peak:
            break
    return left, right


def fit_theo_dist(params, ny, ty):
    right_limit, scaling = params
    index = range(len(ny)) if len(ny) > len(ty) else range(len(ty))
    exp_dist = pd.Series(0, index=index)
    theo_dist = pd.Series(0, index=index)
    exp_dist += pd.Series(ny)
    exp_dist[int(right_limit) :] = 0
    theo_dist += ty
    return ((exp_dist - theo_dist * scaling) ** 2).sum()


def looper(selected=None, df=None, theo=None, index=0, out=None):
    if out is None:
        out = [0] * len(selected)
    if index != len(selected):
        for i in selected[index]:
            out[index] = i
            for j in looper(
                selected=selected, df=df, theo=theo, index=index + 1, out=out
            ):
                yield j
    else:
        vals = pd.Series([df[i] for i in out])
        vals = (vals / vals.max()).fillna(0)
        residual = ((theo - vals) ** 2).sum()
        yield (residual, copy.deepcopy(out))


def find_prior_scan(msn_map, current_scan, ms_level=None):
    prior_msn_scans = {}
    for scan_msn, scan_id in msn_map:
        if scan_id == current_scan:
            return prior_msn_scans.get(
                ms_level if ms_level is not None else scan_msn, None
            )
        prior_msn_scans[scan_msn] = scan_id
    return None


def find_next_scan(msn_map, current_scan, ms_level=None):
    scan_found = False
    for scan_msn, scan_id in msn_map:
        if scan_found:
            if ms_level is None:
                return scan_id
            elif scan_msn == ms_level:
                return scan_id
        if not scan_found and scan_id == current_scan:
            scan_found = True
    return None


def find_scan(msn_map, current_scan):
    for scan_msn, scan_id in msn_map:
        if scan_id == current_scan:
            return scan_id
    return None


def get_scans_under_peaks(rt_scan_map, found_peaks):
    scans = {}
    for peak_isotope, isotope_peak_data in six.iteritems(found_peaks):
        scans[peak_isotope] = {}
        for xic_peak_index, xic_peak_params in six.iteritems(isotope_peak_data):
            mean, stdl, stdr = (
                xic_peak_params["peak_mean"],
                xic_peak_params["std"],
                xic_peak_params["std2"],
            )
            left, right = mean - 2 * stdl, mean + 2 * stdr
            scans[peak_isotope][xic_peak_index] = set(
                rt_scan_map[
                    (rt_scan_map.index >= left) & (rt_scan_map.index <= right)
                ].values
            )
    return scans


def select_window(data, index, size):
    left = int(index - size)
    right = int(index + size)
    if left < 0:
        left = 0
    if right >= len(data) - 1:
        right = -1
    return data[left:] if right == -1 else data[left : right + 1]


def find_common_peak_mean(found_peaks, tie_breaker_time=None):
    # We want to find the peak that is common across multiple scans. This is acheived by the following logic.
    # Suppose we have 3 scans providing peaks. To find our reference, it is likely the peak that occurrs across
    # all scans. A problem that we have to face is the x axis is not guaranteed the same between scans, so we need to
    # establish the overlap of the peak and the peak width. This can be imagined as so:
    #
    # Key: | represent peak bounds, x represents mean. WOO ASCII ART!
    #
    # Scan 1:
    #                             |--x--|                |-x-----|             |-------------x---|
    # Scan 2:
    #                         |---x--|        |---x--|
    #
    # Scan 3:
    #                                 |---x--|                         |----x-----|
    #
    # Thus, the left-most peak is likely the peak we are searching for, and we can select the peaks closest to the mean
    # of these 3 scans as the real peak in our experimental data.
    potential_peaks = {}
    # A peak comparsion of A->B is the same as B->A, so we use combinations to do the minimal amount of work
    # First, we remove a level of nesting to make this easier
    new_peaks = {}
    for peaks_label, peaks_info in six.iteritems(found_peaks):
        for ion, peaks in six.iteritems(peaks_info):
            new_peaks[(peaks_label, ion)] = peaks

    for peaks1_key, peaks2_key in combinations(new_peaks.keys(), 2):
        peaks1 = new_peaks[peaks1_key]
        peaks2 = new_peaks[peaks2_key]
        for peak_index1, peak1 in enumerate(peaks1):
            area, mean = peak1.get("total"), peak1.get("mean")
            left_bound, right_bound = (
                mean - peak1.get("std1", 0) * 2,
                mean + peak1.get("std2", peak1.get("std1", 0)) * 2,
            )
            dict_key = (peaks1_key, peak_index1)
            try:
                potential_peaks[dict_key]["intensities"] += area
            except KeyError:
                potential_peaks[dict_key] = {"intensities": area, "overlaps": set()}
            for peak_index2, peak2 in enumerate(peaks2):
                area2, mean2 = peak2.get("total"), peak2.get("mean")
                left_bound2, right_bound2 = (
                    mean2 - peak2.get("std1", 0) * 2,
                    mean2 + peak2.get("std2", peak2.get("std1", 0)) * 2,
                )
                # This tests whether there is a overlap.
                if (left_bound <= right_bound2) and (left_bound2 <= right_bound):
                    potential_peaks[dict_key]["overlaps"].add(
                        ((peaks2_key, peak_index2), mean2)
                    )
                    potential_peaks[dict_key]["intensities"] += area2
    peak_overlaps = [
        (key, len(overlap_info["overlaps"]), overlap_info["intensities"])
        for key, overlap_info in six.iteritems(potential_peaks)
    ]
    if not peak_overlaps and len(new_peaks) == 1:
        # there is only 1 ion w/ peaks, just pick the biggest peak
        means = [
            sorted(list(new_peaks.values())[0], key=itemgetter("total"), reverse=True)[
                0
            ].get("mean")
        ]
    else:
        sorted_peaks = sorted(peak_overlaps, key=itemgetter(1, 2), reverse=True)
        first_likely_peak = sorted_peaks[0]
        most_likely_peaks = [first_likely_peak]
        if len(sorted_peaks) > 1:
            for likely_peak in sorted_peaks[1:]:
                if likely_peak[1:] == first_likely_peak[1:]:
                    most_likely_peaks.append(likely_peak)
                else:
                    break
        # If we have equi-probable peaks, choose the one that is closest to the tie-breaker time, otherwise return the first
        if tie_breaker_time is not None:
            tie_breaker_distances = [
                (
                    (key, peak_index),
                    np.abs(tie_breaker_time - new_peaks[key][peak_index]["mean"]),
                )
                for (key, peak_index), overlaps, intensity in most_likely_peaks
            ]
            tie_breaker_distances.sort(key=itemgetter(1))
            most_likely_peak = tie_breaker_distances[0][0]
        else:
            most_likely_peak = most_likely_peaks[0][0]

        means = [i[1] for i in potential_peaks[most_likely_peak]["overlaps"]]
        # add in the mean of the initial peak
        means.append(new_peaks[most_likely_peak[0]][most_likely_peak[1]].get("mean"))
    return float(sum(means)) / len(means)


def nanmean(arr, empty=0):
    with warnings.catch_warnings():
        warnings.simplefilter(ERRORS, category=RuntimeWarning)
        if any((i for i in arr if not pd.isnull(i))):
            if isinstance(arr, list):
                return np.nanmean(arr)
            elif arr.any():
                return np.nanmean(arr)
        return empty


def divide_peaks(peaks, min_sep=5, chunk_factor=0.1):
    # We divide up the list of peaks to reduce the number of dimensions each fitting routine is working on
    # to improve convergence speeds. Note -- this function can never rely on our peak estimates because we
    # are very conservative in our peak initial parameter boundaries -- thus we can never really say whether
    # two peaks are going to overlap. This function should instead find dead spots in the data where it is
    # safe to split apart peaks because the overlap would be negligible.

    chunks = argrelextrema(
        np.abs(peaks),
        # To account for peak overlaps, we need to ensure we do not separate peaks where their tails
        # are overlapping by a significant amount
        lambda x, y: np.less(x, y * chunk_factor),
        order=min_sep,
    )[0]
    return chunks


def inffilter(arr):
    with warnings.catch_warnings():
        warnings.simplefilter(ERRORS, category=RuntimeWarning)
        return filter(lambda x: x not in (np.inf, -np.inf), arr)


def naninfmean(arr, empty=0):
    with warnings.catch_warnings():
        warnings.simplefilter(ERRORS, category=RuntimeWarning)
        arr = list(inffilter(arr))
        if any((i for i in arr if not pd.isnull(i))):
            return np.nanmean(arr)
        return empty


def naninfsum(arr, empty=0):
    with warnings.catch_warnings():
        warnings.simplefilter(ERRORS, category=RuntimeWarning)
        arr = list(inffilter(arr))
        if any((i for i in arr if not pd.isnull(i))):
            return np.nansum(arr)
        return empty


ETNS = {
    1: {"C": 0.0110, "H": 0.00015, "N": 0.0037, "O": 0.00038, "S": 0.0075},
    2: {"O": 0.0020, "S": 0.0421},
    4: {"S": 0.00020},
}


def calculate_theoretical_distribution(peptide=None, elemental_composition=None):
    from pythomics.proteomics.config import RESIDUE_COMPOSITION

    def dio_solve(n, loops=None, index=0, out=None):
        loops = loops or [1, 2, 4]
        if out is None:
            out = [0] * len(loops)
        if index != len(loops):
            for i in range(int(n / loops[index]) + 1):
                out[index] = i
                for j in dio_solve(n, loops=loops, index=index + 1, out=out):
                    yield j
        else:
            if n == sum([a * b for a, b in zip(loops, out)]):
                yield out

    ETN_P = {}
    if peptide is not None:
        elemental_composition = {}
        aa_counts = Counter(peptide)
        for aa, aa_count in aa_counts.items():
            for element, element_count in RESIDUE_COMPOSITION[aa].items():
                try:
                    elemental_composition[element] += aa_count * element_count
                except KeyError:
                    elemental_composition[element] = aa_count * element_count
        # we lose a water for every peptide bond
        peptide_bonds = len(peptide) - 1
        elemental_composition["H"] -= peptide_bonds * 2
        elemental_composition["O"] -= peptide_bonds
        # and gain a hydrogen for our NH3
        elemental_composition["H"] += 1

    total_atoms = sum(elemental_composition.values())
    for etn, etn_members in ETNS.items():
        p = 0.0
        for isotope, abundance in etn_members.items():
            p += elemental_composition.get(isotope, 0) * abundance / total_atoms
        ETN_P[etn] = p
    tp = 0
    dist = []
    while tp < 0.999:
        p = 0
        for solution in dio_solve(len(dist)):
            p2 = []
            for k, i in zip(solution, [1, 2, 4]):
                petn = ETN_P[i]
                p2.append(
                    (comb(total_atoms, k) * (petn ** k))
                    * ((1 - petn) ** (total_atoms - k))
                )
            p += np.cumprod(p2)[-1]
        tp += p
        dist.append(p)
    return pd.Series(dist)


def get_classifier():
    import os
    from six.moves.cPickle import load

    pq_dir = os.path.split(__file__)[0]
    handle = open(os.path.join(pq_dir, "static", "classifier.pickle"), "rb")
    classifier = load(handle) if six.PY2 else load(handle, encoding="latin1")

    return classifier


def perform_ml(data, mass_labels):
    import numpy as np
    import sys

    from sklearn.preprocessing import scale
    from scipy.special import logit

    classifier = get_classifier()
    cols = [
        "Isotopes Found",
        "Intensity",
        "RT Width",
        "Mean Offset",
        "Residual",
        "R^2",
        "SNR",
    ]

    for label1 in mass_labels.keys():
        for label2 in mass_labels.keys():
            if label1 == label2:
                continue

            try:
                nd = pd.DataFrame(
                    [],
                    columns=[
                        "Label1 Isotopes Found",
                        "Label1 Intensity",
                        "Label1 RT Width",
                        "Label1 Mean Offset",
                        "Label1 Residual",
                        "Label1 R^2",
                        "Label1 SNR",
                        "Label2 Isotopes Found",
                        "Label2 Intensity",
                        "Label2 RT Width",
                        "Label2 Mean Offset",
                        "Label2 Residual",
                        "Label2 R^2",
                        "Label2 SNR",
                    ],
                )
                for label, new_label in zip([label1, label2], ["Label1", "Label2"]):
                    for col in cols:
                        nd["{} {}".format(new_label, col)] = data[
                            "{} {}".format(label, col)
                        ]

                nd.replace([-np.inf, np.inf, "NA"], np.nan, inplace=True)
                non_na_data = nd.dropna().index
                nd.loc[non_na_data, "Label1 Intensity"] = np.log2(
                    nd.loc[non_na_data, "Label1 Intensity"].astype(float)
                )
                nd.loc[non_na_data, "Label2 Intensity"] = np.log2(
                    nd.loc[non_na_data, "Label2 Intensity"].astype(float)
                )

                nd.replace([-np.inf, np.inf, "NA"], np.nan, inplace=True)
                non_na_data = nd.dropna().index
                nd.loc[non_na_data, "Label1 R^2"] = logit(
                    nd.loc[non_na_data, "Label1 R^2"].astype(float)
                )
                nd.loc[non_na_data, "Label2 R^2"] = logit(
                    nd.loc[non_na_data, "Label2 R^2"].astype(float)
                )

                nd.replace([-np.inf, np.inf, "NA"], np.nan, inplace=True)
                non_na_data = nd.dropna().index
                if not len(non_na_data):
                    continue
                nd.loc[non_na_data, :] = scale(nd.loc[non_na_data, :].values)

                mixed_confidence = "{}/{} Confidence".format(label1, label2)

                # for a 'good' vs. 'bad' classifier, where 1 is good
                conf_ass = classifier.predict_proba(nd.loc[non_na_data, :])[:, 1] * 10
                data.loc[non_na_data, mixed_confidence] = conf_ass

            except Exception as e:
                import traceback

                sys.stderr.write(
                    "Unable to calculate statistics for {}/{}.\n Traceback: {}".format(
                        label1, label2, traceback.format_exc()
                    )
                )


def argrelextrema(data, comparator, axis=0, order=1, mode="clip"):
    results = boolrelextrema(data, comparator, axis, order, mode)
    return np.where(results)


def boolrelextrema(data, comparator, axis=0, order=1, mode="clip"):
    """
    A fixed function for scipy's _boolrelextrama that handles data duplication
    :param data:
    :param comparator:
    :param axis:
    :param order:
    :param mode:
    :return:
    """
    if (int(order) != order) or (order < 1):
        raise ValueError("Order must be an int >= 1")

    datalen = data.shape[axis]
    locs = np.arange(0, datalen)
    data = np.ma.masked_array(data, mask=np.hstack(([1], np.diff(data))) == 0)

    if np.ma.is_masked(data):
        locs = locs[~np.ma.getmask(data)]
        main = data.take(locs, axis=axis, mode=mode)
        results = np.zeros(data.shape, dtype=bool)
        for index, result in enumerate(
            boolrelextrema(main, comparator, axis=axis, order=order, mode=mode)
        ):
            results[locs[index]] = result
        return results
    else:
        results = np.ones(data.shape, dtype=bool)
        main = data.take(locs, axis=axis, mode=mode)
        for shift in range(1, order + 1):
            plus = data.take(locs + shift, axis=axis, mode=mode)
            minus = data.take(locs - shift, axis=axis, mode=mode)
            results &= comparator(main, plus)
            results &= comparator(main, minus)
            if ~results.any():
                return results
        return results


def merge_peaks(peaks_found, debug=False):
    if len(peaks_found) == 1:
        return peaks_found

    final_peaks = {}
    peak_widths = sorted(peaks_found.keys())
    for peak_width_index in range(len(peak_widths) - 1):
        current_width = peak_widths[peak_width_index]
        next_width = peak_widths[peak_width_index + 1]
        current_peaks = peaks_found[current_width]
        next_peaks = peaks_found[next_width]
        smaller_peaks, smaller_minima = current_peaks["peaks"], current_peaks["minima"]
        larger_peaks, larger_minima = next_peaks["peaks"], next_peaks["minima"]
        if debug:
            sys.stderr.write(
                "{}: {} ---- {}\n".format(peak_width_index, smaller_peaks, larger_peaks)
            )
        if set(smaller_peaks) == set(larger_peaks) and set(smaller_minima) == set(
            larger_minima
        ):
            final_peaks[next_width] = next_peaks
            if current_width in final_peaks:
                del final_peaks[current_width]
        else:
            final_peaks[current_width] = peaks_found[current_width]
            if current_width == peak_widths[-2]:
                final_peaks[next_width] = peaks_found[next_width]

    return final_peaks


def find_possible_peaks(
    xdata,
    ydata,
    ydata_peaks,
    peak_find_method=PEAK_FINDING_REL_MAX,
    min_dist=None,
    local_filter_size=0,
    rt_peak=None,
    peak_width_start=2,
    snr=0,
    zscore=0,
    amplitude_filter=0,
    peak_width_end=4,
    fit_negative=False,
    percentile_filter=0,
    micro=False,
    min_slope=None,
    min_peak_side_width=None,
    min_peak_width=5,
    smooth=False,
    min_peak_increase=1.05,
):
    PEAK_METHODS = {
        PEAK_FINDING_REL_MAX: partial(
            find_peaks_rel_max,
            peak_width_start=peak_width_start,
            peak_width_end=peak_width_end,
            micro=micro,
        ),
        PEAK_FINDING_DERIVATIVE: partial(
            find_peaks_derivative,
            min_slope=min_slope,
            min_peak_side_width=min_peak_side_width,
            min_peak_width=min_peak_width,
            smooth=smooth,
            rel_peak_height=min_peak_increase,
        ),
    }
    abs_ydata = np.abs(ydata)
    possible_peaks = PEAK_METHODS[peak_find_method](
        xdata, ydata, ydata_peaks=ydata_peaks
    )

    for peak_width, peak_info in six.iteritems(possible_peaks):
        row_peaks, minima = peak_info["peaks"], peak_info["minima"]
        if min_dist:
            row_peaks = merge_close_peaks(row_peaks, ydata, distance=min_dist)
        if snr or zscore:
            ydata_peaks_std = np.std(ydata_peaks)
            ydata_peaks_median = np.median(ydata_peaks)
            if local_filter_size:
                new_peaks = []
                lost_peaks = {}
                for row_peak in row_peaks:
                    selection = np.abs(
                        select_window(ydata_peaks, row_peak, local_filter_size)
                    )
                    ydata_row_peak = np.abs(ydata_peaks[row_peak])
                    local_std = np.std(selection)
                    local_snr = ydata_row_peak / local_std
                    local_zscore = (ydata_row_peak - np.median(selection)) / local_std
                    add_peak = (snr == 0 or local_snr > snr) and (
                        zscore == 0 or local_zscore >= zscore
                    )
                    if add_peak:
                        new_peaks.append(row_peak)
                    else:
                        lost_peaks[row_peak] = {"snr": local_snr, "zs": local_zscore}
                logger.debug(
                    "{} peaks lost to filtering\n{}\n".format(
                        len(row_peaks) - len(new_peaks), lost_peaks
                    )
                )
                row_peaks = np.array(new_peaks, dtype=int)
            else:
                if snr and row_peaks.any():
                    logger.debug(
                        "{} peaks lost to SNR\n".format(
                            sum(ydata_peaks[row_peaks] / ydata_peaks_std < snr)
                        )
                    )
                if zscore and row_peaks.any():
                    logger.debug(
                        "{} peaks lost to zscore\n".format(
                            sum(
                                (ydata_peaks[row_peaks] - ydata_peaks_median)
                                / ydata_peaks_std
                                < zscore
                            )
                        )
                    )
                if snr and row_peaks.any():
                    row_peaks = row_peaks[
                        np.abs(ydata_peaks[row_peaks]) / ydata_peaks_std >= snr
                    ]
                if zscore and row_peaks.any():
                    row_peaks = row_peaks[
                        (np.abs(ydata_peaks[row_peaks]) - ydata_peaks_median)
                        / ydata_peaks_std
                        >= zscore
                    ]

        if amplitude_filter != 0 and row_peaks.any():
            logger.debug(
                "{} peaks lost to amp filter\n{}\n".format(
                    sum(abs_ydata[row_peaks] < amplitude_filter),
                    row_peaks[abs_ydata[row_peaks] < amplitude_filter],
                )
            )
            row_peaks = row_peaks[np.abs(ydata_peaks[row_peaks]) >= amplitude_filter]

        if percentile_filter and row_peaks.any():
            logger.debug(
                "{} peaks lost to percentile filter\n{}\n".format(
                    sum(
                        np.abs(ydata_peaks[row_peaks])
                        < np.percentile(abs_ydata, percentile_filter)
                    ),
                    row_peaks[
                        np.abs(ydata_peaks[row_peaks])
                        < np.percentile(abs_ydata, percentile_filter)
                    ],
                )
            )
            row_peaks = row_peaks[
                np.abs(ydata_peaks[row_peaks])
                >= np.percentile(abs_ydata, percentile_filter)
            ]

        possible_peaks[peak_width] = {"peaks": row_peaks, "minima": minima}

    return merge_peaks(possible_peaks)


def find_peaks_rel_max(
    xdata, ydata, ydata_peaks=None, peak_width_start=2, peak_width_end=4, micro=False
):
    ydata_peaks = ydata_peaks if ydata_peaks is not None else ydata
    if peak_width_start > peak_width_end:
        peak_width_end = peak_width_start + 1
    peak_width = peak_width_start
    peaks_found = {}
    while peak_width <= peak_width_end:
        row_peaks = np.array(
            argrelextrema(np.abs(ydata_peaks), np.greater, order=peak_width)[0],
            dtype=int,
        )
        if not row_peaks.size:
            row_peaks = np.array([np.argmax(np.abs(ydata))], dtype=int)
        logger.debug("peak indices: {}\n".format(row_peaks))

        if ydata_peaks.size:
            minima = np.where(ydata_peaks == 0)[0].tolist()
        else:
            minima = []
        minima.extend(
            [
                i
                for i in argrelextrema(np.abs(ydata_peaks), np.less, order=peak_width)[
                    0
                ]
                if i not in minima and i not in row_peaks
            ]
        )
        minima.sort()

        peaks_found[peak_width] = {"peaks": row_peaks, "minima": minima}

        peaks_found[peak_width] = {"peaks": row_peaks, "minima": minima}
        peak_width += 1

    return peaks_found


def get_cross_points(data, pad=True):
    cross_points = []
    signs = np.sign(data)
    for i in range(len(signs) - 1):
        if signs[i] != signs[i + 1]:
            cross_points.append(i)

    if cross_points and pad:
        # Pad the ends to deal with the first and last peaks found by prepending the crossing points array
        # with the data point that has the closest opposite sign
        if cross_points[0] != 0:
            cross_points.insert(
                0, np.argmax(signs[: cross_points[0]] != signs[cross_points[0]])
            )

        if cross_points[-1] != len(signs) - 1:
            not_equal = signs[cross_points[-1] + 1 :] != signs[cross_points[-1] + 1]
            if not_equal.any():
                cross_points.append(cross_points[-1] + np.argmax(not_equal) + 1)
            else:
                cross_points.append(len(signs) - 1)

    return cross_points


def find_peaks_derivative(
    xdata,
    ydata,
    ydata_peaks=None,
    min_slope=None,
    rel_peak_height=None,
    min_peak_side_width=2,
    max_peak_side_width=np.inf,
    min_peak_width=None,
    max_peak_width=np.inf,
    smooth=True,
):

    # The general strategy here is to identify where the derivative crosses the zero point, and
    # pick out when the sign of the derivative changes in a manner consistent with a peak.
    # A peak using this approach would like look (pw=peak width):
    #
    #          |     left p.w   |  right p.w.  |
    #    ++--+-++++++++++++++++++--------------+++++++++++
    #                           ^ peak max
    #
    #
    rel_peak_height = rel_peak_height if rel_peak_height is not None else 1.05
    peaks_found = {"peaks": [], "minima": []}
    ydata_peaks = ydata_peaks if ydata_peaks is not None else ydata
    ydata = np.abs(ydata_peaks)

    if min_peak_width is None:
        min_peak_width = int(len(ydata) / 2)
        if min_peak_width > 5:
            min_peak_width = 5

    # We first take the derivative of the data and smooth the derivative to reduce noise
    deriv = np.diff(ydata)
    smoothed_deriv = savgol_smooth(deriv) if smooth else deriv
    cross_points = get_cross_points(smoothed_deriv)
    # Because we take the difference, we need to figure out whether the left or right is the true peak for crosses.
    # By default, cross points returns the left side
    for i in range(len(cross_points)):
        index = cross_points[i]
        if index < len(ydata):
            if ydata[index] < ydata[index + 1]:
                cross_points[i] = index + 1

    if cross_points:
        # Next, we compare crossing points to identify peaks and constrain it using some criteria:
        # slope, peak width
        for i in range(1, len(cross_points) - 1):
            left, center, right = (
                cross_points[i - 1],
                cross_points[i],
                cross_points[i + 1],
            )
            left_x, center_x, right_x = xdata[left], xdata[center], xdata[right]
            left_y, center_y, right_y = ydata[left], ydata[center], ydata[right]

            logger.debug(
                "Looking at peak %s",
                (
                    left,
                    center,
                    right,
                    left_x,
                    center_x,
                    right_x,
                    left_y,
                    center_y,
                    right_y,
                ),
            )

            # Most basic assumption -- does it look like a peak?
            if not left_y < center_y > right_y:
                logger.debug("%s doesnt look like a peak", center_x)
                continue

            # Does it meet our peak width criteria?
            left_peak_width = center - left
            right_peak_width = right - center
            if not (
                (
                    (min_peak_side_width <= left_peak_width <= max_peak_side_width)
                    or (
                        min_peak_side_width > left
                    )  # This allows us to find peaks that start at the edges of our data
                )
                and (
                    (min_peak_side_width <= right_peak_width <= max_peak_side_width)
                    or (
                        (len(ydata) - 1) < right + min_peak_side_width
                    )  # This allows us to find peaks that start at the edges of our data
                )
            ):
                logger.debug("peak half not wide enough")
                continue

            peak_width = right - left
            if not (min_peak_width < peak_width < max_peak_width):
                logger.debug("peak not wide enough")
                continue

            slope = max(
                np.abs(
                    [
                        (center_y - left_y) / (center_x - left_x),
                        (center_y - right_y) / (center_x - right_x),
                    ]
                )
            )

            # Both sides of the peak shape must increase beyond a given threshold so we take the minima here
            rel_peak = min(np.abs([(center_y / left_y), (center_y / right_y),]))

            if min_slope is not None and slope < min_slope:
                logger.debug("not sloped enough")
                continue
            if rel_peak < rel_peak_height:
                logger.debug("not tall enough")
                continue

            peaks_found["peaks"].append(center)
            peaks_found["minima"].append(left)
            peaks_found["minima"].append(right)

    peaks_found["peaks"] = np.array(peaks_found["peaks"])
    peaks_found["minima"] = np.array(peaks_found["minima"])

    # We return our peaks like:
    # { peak_search_width: {'peaks': [], 'minima': []}} to be consistent with other peak finding routines
    return {min_peak_width: peaks_found}


def erode_direction(peaks_found, peak_heights, start, increment, distance):
    peak = peaks_found[start]
    peak_height = peak_heights[start]
    next_index = start + increment
    peaks_to_remove = set([])
    while (
        (next_index >= 0 and next_index < len(peaks_found))
        and np.abs(peak - peaks_found[next_index]) < distance
        and peak_height != peak_heights[next_index]
    ):
        peaks_to_remove.add(peaks_found[next_index])
        next_index += increment
    return peaks_to_remove


def merge_close_peaks(peaks_found, ydata, distance=2):
    new_peaks = []
    removed_peaks = set([])
    if peaks_found.any():
        peak_heights = ydata[peaks_found]
        sort_order = np.argsort(peak_heights)[::-1]

        for sort_index, peak_index in enumerate(sort_order):
            peak = peaks_found[peak_index]
            if peak in removed_peaks:
                continue
            new_peaks.append(peak_index)

            removed_peaks |= erode_direction(
                peaks_found, peak_heights, peak_index, 1, distance
            )
            removed_peaks |= erode_direction(
                peaks_found, peak_heights, peak_index, -1, distance
            )

    return np.array([i for i in peaks_found if i not in removed_peaks])


def estimate_peak_parameters(
    xdata,
    ydata,
    row_peaks,
    minima_array,
    fit_negative=False,
    rel_peak_constraint=0,
    micro=False,
    bigauss_fit=False,
    fit_baseline=False,
):
    guess = []
    bnds = []
    last_peak = -1
    skip_peaks = set([])
    fitted_peaks = []
    min_spacing = min(np.diff(xdata)) / 2
    peak_range = xdata[-1] - xdata[0]
    for peak_num, peak_index in enumerate(row_peaks):
        if peak_index in skip_peaks:
            continue
        next_peak = (
            len(xdata) if peak_index == row_peaks[-1] else row_peaks[peak_num + 1]
        )
        fitted_peaks.append(peak_index)
        rel_peak = ydata[peak_index]
        # bounds for fitting the peak mean
        peak_left = xdata[peak_index - 1]
        peak_right = xdata[peak_index + 1] if peak_index + 1 < len(xdata) else xdata[-1]
        # find the points around it to estimate the std of the peak
        if minima_array.size:
            left = np.searchsorted(minima_array, peak_index) - 1
            left_stop = (
                np.searchsorted(minima_array, last_peak) if last_peak != -1 else -1
            )
            if left == -1:
                left = 0 if last_peak != -1 else last_peak + 1
            elif left == left_stop:
                if last_peak == -1:
                    left = 0
                else:
                    left = minima_array[left]
            elif left_stop > left:
                # We are at the last minima, set our left bound to the last peak if it is greater than
                # the left minima, otherwise set to the left minima
                minima_index = minima_array[left]
                left = last_peak if last_peak > minima_index else minima_index
            else:
                left = minima_array[left]
            last_peak = peak_index
            right = np.searchsorted(minima_array, peak_index)
            if right >= minima_array.size:
                right = minima_array[-1]
            else:
                try:
                    right = minima_array[right]
                except Exception as e:
                    print(right, ydata, minima_array, peak_index)
            if right > next_peak:
                right = next_peak
            if right < peak_index:
                right = next_peak
            if right >= len(xdata) - 1:
                right = len(xdata) - 1
            bnds.extend(
                [
                    (-1.01, rel_peak * rel_peak_constraint)
                    if rel_peak < 0 and fit_negative
                    else (rel_peak * rel_peak_constraint, 1.01),
                    (peak_left, peak_right) if micro else (xdata[left], xdata[right]),
                    (min_spacing, peak_range),
                ]
            )
            if bigauss_fit:
                bnds.extend([(min_spacing, peak_range)])
            peak_values = ydata[left:right]
            peak_indices = xdata[left:right]
        else:
            left = 0
            right = len(xdata) - 1
            bnds.extend(
                [
                    (-1.01, rel_peak * rel_peak_constraint)
                    if rel_peak < 0 and fit_negative
                    else (rel_peak * rel_peak_constraint, 1.01),
                    (peak_left, peak_right) if micro else (xdata[0], xdata[-1]),
                    (min_spacing, peak_range),
                ]
            )
            if bigauss_fit:
                bnds.extend([(min_spacing, peak_range)])
            peak_values = ydata[left : right + 1]
            peak_indices = xdata[left : right + 1]

        logger.debug(
            "bounds are {}".format(
                (
                    peak_index,
                    left,
                    right,
                    peak_values.tolist(),
                    peak_indices.tolist(),
                    bnds,
                )
            )
        )

        if peak_values.any():
            average = np.average(peak_indices, weights=np.abs(peak_values))
            variance = np.sqrt(
                np.average((peak_indices - average) ** 2, weights=np.abs(peak_values))
            )
            if variance == 0:
                # we have a singular peak if variance == 0, so set the variance to half of the x/y spacing
                if peak_index >= 1:
                    variance = np.abs(xdata[peak_index] - xdata[peak_index - 1])
                elif peak_index < len(xdata):
                    variance = np.abs(xdata[peak_index + 1] - xdata[peak_index])
                else:
                    # we have only 1 data point, most RT's fall into this width
                    variance = 0.05
        else:
            variance = 0.05
            average = xdata[peak_index]
        if variance is not None and variance < min_spacing:
            variance = min_spacing
        if variance is not None:
            if bigauss_fit:
                guess.extend([rel_peak, xdata[peak_index], variance, variance])
            else:
                guess.extend([rel_peak, xdata[peak_index], variance])
            if fit_baseline:
                slope = (ydata[right] - ydata[left]) / (xdata[right] - xdata[left])
                intercept = (
                    (ydata[right] - slope * xdata[right])
                    + (ydata[left] - slope * xdata[left])
                ) / 2
                guess.extend([slope, intercept])
                bnds.extend([(None, None), (None, None)])

    if not guess:
        average = np.average(xdata, weights=np.abs(ydata))
        variance = np.sqrt(np.average((xdata - average) ** 2, weights=np.abs(ydata)))
        if variance == 0:
            variance = 0.05
        guess = [max(ydata), average, variance]
        if bigauss_fit:
            guess.extend([variance])
        if fit_baseline:
            slope = (ydata[-1] - ydata[0]) / (xdata[-1] - xdata[0])
            intercept = (
                (ydata[-1] - slope * xdata[-1]) + (ydata[0] - slope * xdata[0])
            ) / 2
            guess.extend([slope, intercept])
            bnds.extend([(None, None), (None, None)])

    if not bnds:
        bnds.extend([(None, None) for i in guess])

    return guess, bnds


def interpolate_data(x, y, gap_limit=2):
    from scipy.interpolate import interp1d

    x, y = np.array(x), np.array(y)
    pos_indices = np.where(y > 0)[0]
    if pos_indices.size >= 2:
        mapper = interp1d(x[pos_indices], y[pos_indices])
        gaps = np.diff(pos_indices)
        to_fill = np.where((gaps > 1) & (gaps <= gap_limit + 1))[0]
        for index in to_fill:
            left = pos_indices[index] + 1
            right = pos_indices[index + 1]
            y[left:right] = mapper(x[left:right])
    return y


def savgol_smooth(ydata, max_window=10, polyorder=3):
    window_size = int(len(ydata) / 10)
    if window_size < 5:
        window_size = 5
    elif window_size > max_window:
        window_size = max_window

    if window_size > len(ydata):
        window_size = len(ydata)

    if not window_size % 2:
        window_size -= 1

    if polyorder >= window_size:
        polyorder = window_size - 1

    return savgol_filter(ydata, window_size, polyorder)


def get_formatted_mass(mass):
    """
    This helps with looking up masses in a dictionary by mass by ensuring the values 0.3 and 0.30 are represented
    the same
    :param mass:
    :return:
    """
    return float("{0:0.6f}".format(float(mass)))


def get_scan_resolution(scan):
    """

    :param scan: A scan series (pandas series)
    :return:
    """
    from .peaks import findAllPeaks

    max_peak = np.argmax(scan.values)

    left = max_peak - 30
    right = max_peak + 30
    if left < 0:
        left = 0
    if right > len(scan):
        right = len(scan)

    x = scan.index.values[left:right].astype(float)
    y = scan.values[left:right].astype(float)
    peaks, residual = findAllPeaks(x, y, peak_width_start=1, max_peaks=-1)
    if peaks is None or not peaks.any():
        print("no peaks in ", scan)
        return 0

    largest_peak = np.argmax(peaks[0::3])

    fwhm = peaks[2 + largest_peak * 3] * 2.355
    return peaks[1 + largest_peak * 3] / fwhm


def subtract_baseline(y, components_to_remove=10, dampen_factor=10):
    from scipy import fftpack, signal

    f = fftpack.rfft(y)
    f[:components_to_remove] = 0
    fs = fftpack.fftshift(f)

    # filter it
    filtered = (
        (fs * (1 - signal.gaussian(len(y), dampen_factor))) if dampen_factor else fs
    )
    # invert it
    final = fftpack.irfft(fftpack.ifftshift(filtered))

    return final



__author__ = "chris"
import os
import unittest

import numpy as np
import pickle

from .utils import timer
from .mixins import FileMixins, GaussianMixin
from pyquant import peaks
from pyquant import PEAK_FINDING_DERIVATIVE, PEAK_FINDING_REL_MAX, PEAK_FIT_MODE_FAST


def get_gauss_value(x, amp, mu, std):
    return amp * np.exp(-((x - mu) ** 2) / (2 * std ** 2))


class PeakFindingTests(FileMixins, unittest.TestCase):
    def test_returns_with_bad_data(self):
        params, residual = peaks.findAllPeaks(
            np.array([1, 2, 3, 4, 5, 6]), np.array([0, 0, 0, 0, 0, 0])
        )
        self.assertEqual(len(params), 0)
        self.assertEqual(residual, np.inf)

    def test_max_peaks(self):
        # Regression where relative-max is reporting 2 peaks when max_peaks is set to 1. This occurred
        # because we enforced max_peaks for each peak width when using the relative-max setting. Thus,
        # the max peak for each peak width was combined to the final peak report. The update was to
        # pick the peak_width with the lowest BIC.
        with open(os.path.join(self.data_dir, "peak_data.pickle"), "rb") as peak_file:
            data = pickle.load(peak_file, encoding="latin1")

        x, y = data["max_peaks_relative-max"]
        params, residual = peaks.findAllPeaks(
            x,
            y,
            max_peaks=1,
            peak_find_method=PEAK_FINDING_REL_MAX,
            fit_mode=PEAK_FIT_MODE_FAST,
        )
        self.assertEqual(len(params), 3)

    def test_max_peaks_with_rt_peak_regression(self):
        with open(os.path.join(self.data_dir, "peak_data.pickle"), "rb") as peak_file:
            data = pickle.load(peak_file, encoding="latin1")

        x, y = data["max_peaks_rt-peak-regression"]
        params, residual = peaks.findAllPeaks(
            x, y, max_peaks=1, rt_peak=360, fit_mode=PEAK_FIT_MODE_FAST
        )
        np.testing.assert_allclose(params[1], desired=365.78, atol=0.1)

    def test_fit_baseline_derivative(self):
        with open(os.path.join(self.data_dir, "peak_data.pickle"), "rb") as peak_file:
            data = pickle.load(peak_file, encoding="latin1")

        x, y = data["max_peaks_rt-peak-regression"]
        params, residual = peaks.findAllPeaks(
            x,
            y,
            max_peaks=1,
            fit_baseline=True,
            rt_peak=328,
            peak_find_method="derivative",
            fit_mode=PEAK_FIT_MODE_FAST,
        )
        np.testing.assert_allclose(
            params[:3], desired=np.array([1320.60, 330.15, 4.22]), atol=10
        )

    def test_segmenty_negatives(self):
        # Regression where a mostly positive dataset with negatives led to -inf values in the data array
        # due to np.max(segment_y) being 0 since all data was negative
        with open(os.path.join(self.data_dir, "peak_data.pickle"), "rb") as peak_file:
            data = pickle.load(peak_file, encoding="latin1")

        x, y = data["invalid_operands"]
        params, res = peaks.findAllPeaks(
            x,
            y,
            max_peaks=-1,
            bigauss_fit=True,
            peak_find_method=PEAK_FINDING_DERIVATIVE,
        )
        means = params[1::4]
        desired = np.array(
            [
                0.13515435795212014,
                0.33,
                1.474992882679938,
                1.799090776628427,
                2.1804381077669395,
                2.6350000000000002,
                3.227084689771589,
                3.617021549048893,
                4.903333333333333,
                5.296162908137783,
                5.8366172292356175,
            ]
        )
        np.testing.assert_allclose(means, desired=desired, atol=0.1)


class GaussianTests(GaussianMixin, unittest.TestCase):
    @timer
    def test_gauss_ndim(self):
        assert np.round(self.two_gauss[np.where(self.x == 0)], 2) == np.round(
            get_gauss_value(0, self.amp, self.mu, self.std)
            + get_gauss_value(0, self.amp, self.mu2, self.std),
            2,
        )

    @timer
    def test_peak_fitting(self):
        # first, a simple case
        params, residual = peaks.findAllPeaks(self.x, self.one_gauss,)
        np.testing.assert_allclose(params, self.one_gauss_params, atol=self.std / 2)
        params, residual = peaks.findAllPeaks(
            self.x, self.one_gauss, peak_find_method=PEAK_FINDING_DERIVATIVE
        )
        np.testing.assert_allclose(params, self.one_gauss_params, atol=self.std / 2)

        params, residual = peaks.findAllPeaks(self.x, self.two_gauss)
        np.testing.assert_allclose(params, self.two_gauss_params, atol=self.std / 2)
        params, residual = peaks.findAllPeaks(
            self.x, self.two_gauss, peak_find_method=PEAK_FINDING_DERIVATIVE
        )
        np.testing.assert_allclose(params, self.two_gauss_params, atol=self.std / 2)
        params, residual = peaks.findAllPeaks(
            self.x,
            self.noisy_two_gauss,
            rt_peak=None,
            filter=True,
            max_peaks=30,
            bigauss_fit=True,
            snr=1,
        )
        np.testing.assert_allclose(
            params[1::4], self.two_gauss_params[1::3], atol=self.std / 2
        )

    def test_targeted_search(self):
        # We should not find anything where there are no peaks
        res, residual = peaks.targeted_search(
            self.x,
            self.two_gauss,
            self.x[2],
            attempts=2,
            peak_finding_kwargs={"max_peaks": 2},
        )
        self.assertIsNone(res)

        # Should find the peak when we are in its area
        res, residual = peaks.targeted_search(
            self.x, self.two_gauss, self.two_gauss_params[1]
        )
        self.assertIsNotNone(res)

    def test_experimental(self):
        # Experimental data
        x, y = self.peak_data["offset_fit"]
        params, residual = peaks.findAllPeaks(x, y, bigauss_fit=True, filter=True)
        np.testing.assert_allclose(
            params,
            np.array(
                [
                    13219262.587656807,
                    46.821340819991505,
                    0.06523272363478014,
                    0.18374422913588656,
                    3347200.309180678,
                    47.497,
                    0.6166821402545103,
                    0.3817876338966981,
                    1880722.1582992678,
                    48.14756707645761,
                    0.17537885391522443,
                    0.4846763157315077,
                    1473766.5256005626,
                    49.52607160264086,
                    0.020199999999999108,
                    0.22250905781532157,
                ]
            ),
            # The rtol is 1e-7, so we really don't care about the atol
            atol=20,
        )


if __name__ == "__main__":
    unittest.main()



import os
import subprocess
import unittest

import pandas as pd
import numpy as np

import tempfile

from .mixins import FileMixins
from . import config


class TestTargeted(FileMixins, unittest.TestCase):
    def setUp(self):
        super(TestTargeted, self).setUp()
        self.output = os.path.join(self.out_dir, "targeted")

    def test_pyquant_trypsin(self):
        # This searches for the y1 ions of arginine. It also is a check that the label-scheme parameter works.
        f = tempfile.NamedTemporaryFile("w")
        f.write("\t".join(["0", "R", "10.008269", "R10"]) + "\n")
        f.write("\t".join(["1", "R", "0", "Light"]) + "\n")
        f.seek(0)
        com = [
            self.executable,
            "--scan-file",
            self.ecoli_mzml,
            "-p",
            str(config.CORES),
            "-o",
            self.output,
            "--msn-ion",
            "175",
            "--html",
            "--label-scheme",
            f.name,
        ]
        subprocess.call(com)
        pyquant = pd.read_table(self.output)
        label = "R10"
        pq_sel = "{}/Light".format(label)
        pyquant[pq_sel] = np.log2(pyquant[pq_sel] + 0.000001)
        r10_med = pyquant[pq_sel].median()
        self.assertNotAlmostEqual(r10_med, -2.0, places=2)


if __name__ == "__main__":
    unittest.main()



__author__ = "chris"
import os
import unittest
import pandas as pd
import numpy as np
import subprocess

from . import mixins, utils, config


class EColiTest(mixins.FileMixins, unittest.TestCase):
    def setUp(self):
        super(EColiTest, self).setUp()
        self.output = os.path.join(self.out_dir, "pqtest2")
        self.output_stats = "{}_stats".format(self.output)
        self.r_std = 0.6
        self.k_std = 0.9

    @utils.timer
    def test_pyquant(self):
        com = [
            self.executable,
            "--search-file",
            self.ecoli_search_file,
            "--scan-file",
            self.ecoli_mzml,
            "-p",
            str(config.CORES),
            "-o",
            self.output,
            "--html",
            "--precursor-ppm",
            "2",
            "--xic-window-size",
            "12",
        ]
        subprocess.call(com)
        pyquant = pd.read_table(self.output)
        label = "Medium"
        pq_sel = "{}/Light".format(label)
        pyquant.loc[
            (
                (pyquant["Peptide"].str.upper().str.count("R") == 1)
                & (pyquant["Peptide"].str.upper().str.count("K") == 0)
            ),
            "Class",
        ] = "R"
        pyquant.loc[
            (
                (pyquant["Peptide"].str.upper().str.count("K") == 1)
                & (pyquant["Peptide"].str.upper().str.count("R") == 0)
            ),
            "Class",
        ] = "K"
        pyquant[pq_sel].apply(lambda x: np.log2(x) if x != 0 else x)
        # Mark zeros as nan to exclude from calculation
        pyquant.loc[pyquant[pq_sel] == 0, pq_sel] = np.NaN
        # the median is robust, we care about the standard deviation since changes to the backend can alter the peak width
        r_std = np.std(pyquant.loc[pyquant["Class"] == "R", pq_sel])
        k_std = np.std(pyquant.loc[pyquant["Class"] == "K", pq_sel])
        self.assertLess(r_std, self.r_std)
        self.assertLess(k_std, self.k_std)


if __name__ == "__main__":
    unittest.main()
